from pathlib import Path
import pytest
import re

from datalad_next.consts import on_windows
from datalad_next.tests import (
    create_tree,
    skip_if_on_windows,
    skip_if_root,
)
from datalad_next.constraints import EnsureDataset
from datalad_next.exceptions import (
    CommandError,
    UrlOperationsRemoteError,
    IncompleteResultsError,
)
from datalad_next.url_operations import AnyUrlOperations

from ..uncurl import (
    RemoteError,
    UncurlRemote,
)

# for some tests below it is important that this base config contains no
# url= or match= declaration (or any other tailoring to a specific use case)
std_initargs = [
    'type=external',
    'externaltype=uncurl',
    'encryption=none',
]


class NoOpAnnex:
    def __init__(self, gitdir: Path):
        self._gitdir = gitdir

    def error(*args, **kwargs):
        pass

    def info(*args, **kwargs):
        pass

    def debug(*args, **kwargs):
        pass

    def getgitdir(self):
        return str(self._gitdir)

    def getgitremotename(self):
        return 'origin'

    def getconfig(self, key):
        return None

    def getuuid(self):
        return '2d0660ae-e5bf-11ee-b6f2-37ba632b6dd3'


def test_uncurl_remove_no_tmpl(existing_dataset):
    # without a template configured we refuse to remove anything
    # for the simple reason that it may not be clear what is being
    # removed at all. We could only iterate over all recorded URLs
    # and wipe out the key from "the internet". This is, however,
    # a rather unexpected thing to do from a user perspective --
    # who would expect a single key instance "at the uncurl remote"
    # to be removed. The best proxy we have for this expectation
    # is a URL tmpl being configured, pointing to such a single
    # location
    r = UncurlRemote(NoOpAnnex(existing_dataset.pathobj))
    r.prepare()
    with pytest.raises(RemoteError):
        r.remove(None)


def test_uncurl_transfer_store_no_tmpl(existing_dataset):
    r = UncurlRemote(NoOpAnnex(existing_dataset.pathobj))
    r.url_handler = AnyUrlOperations()
    # whenever there is not template configured
    with pytest.raises(RemoteError):
        r.transfer_store(None, '')


def test_uncurl_checkretrieve(existing_dataset):

    def handler(url):
        raise UrlOperationsRemoteError(url)

    def get_urls(key):
        return 'some'

    r = UncurlRemote(NoOpAnnex(existing_dataset.pathobj))
    r.prepare()
    r.get_key_urls = get_urls

    # we raise the correct RemoteError and not the underlying
    # UrlOperationsRemoteError
    with pytest.raises(RemoteError):
        r._check_retrieve('somekey', handler, ('blow', 'here'))


def test_uncurl_claimurl(existing_dataset):
    r = UncurlRemote(NoOpAnnex(existing_dataset.pathobj))
    r.prepare()
    # if we have a match expression defined, this determines claim or noclaim
    r.match = [re.compile('bongo.*')]
    assert r.claimurl('bongo://joe')
    assert not r.claimurl('http://example.com')
    r.match = None
    # without a match expression, the url handler decides
    r.url_handler = AnyUrlOperations()
    for url in ('http://example.com',
                'https://example.com',
                'ssh://example.com',
                'file:///home/me'):
        assert r.claimurl(url)
    assert not r.claimurl('bongo://joe')


def test_uncurl_checkurl(httpbin, tmp_path, existing_dataset):
    # this is the URL against which the httpbin calls will be made
    hbsurl = httpbin['standard']

    exists_path = tmp_path / 'testfile'
    exists_path.write_text('123')
    exists_url = exists_path.as_uri()
    no_exists_url = (tmp_path / 'notestfile').as_uri()
    # checkurl is basically an 'exists?' test against a URL.
    # the catch is that this test is not performed against the
    # incoming URL, but against the mangled URL that is the result
    # of instantiating the URL template based on all properties
    # extractable from the URL alone (via any configured match
    # expressions)
    r = UncurlRemote(NoOpAnnex(existing_dataset.pathobj))
    r.prepare()
    # no match and no template defined
    assert not r.checkurl(no_exists_url)
    assert r.checkurl(exists_url)
    #
    # now with rewriting
    #
    # MIH cannot think of a usecase where declaring a URL template
    # would serve any purpose without also declaring a match expression
    # here.
    # outside the context of checkurl() this is different: It would
    # make sense to declare a template that only uses standard key properties
    # in order to define/declare upload targets for existing keys.
    # consequently, checkurl() is ignoring a template-based rewriting
    # when no match is defined, or when the matchers cannot extract all
    # necessary identifiers from the incoming (single) URL in order to
    # instantiate the URL template. In such cases, the original URL is
    # used for checking
    r.url_tmpl = '{absurd}'
    assert not r.checkurl(no_exists_url)
    assert r.checkurl(exists_url)

    # now add a matcher to make use of URL rewriting even for 'addurl'-type
    # use case, such as: we want to pass a "fixed" URL verbatim to some kind
    # of external redirector service
    r.url_tmpl = f'{hbsurl}/redirect-to?url={{origurl}}'
    r.match = [
        re.compile('.*(?P<origurl>https?://.*)$'),
    ]
    assert not r.checkurl(f'garbled{hbsurl}/status/404')
    assert r.checkurl(f'garbled{hbsurl}/bytes/24')


# sibling of `test_uncurl_checkurl()`, but more high-level
def test_uncurl_addurl_unredirected(
        existing_dataset, httpbin, no_result_rendering):
    # this is the URL against which the httpbin calls will be made
    hbsurl = httpbin['standard']

    ds = existing_dataset
    dsca = ds.repo.call_annex
    # same set as in `test_uncurl_checkurl()`
    dsca(['initremote', 'myuncurl'] + std_initargs + [
        'match=.*(?P<origurl>https?://.*)$',
        f'url={hbsurl}/redirect-to?url={{origurl}}',
    ])
    # feed it a broken URL, which must be getting fixed by the rewriting
    # (pulls 24 bytes)
    testurl = f'garbled{hbsurl}/bytes/24'
    dsca(['addurl', '--file=dummy', testurl])
    # we got what we expected
    assert ds.status(
        'dummy', annex='basic', return_type='item-or-list',
        )['bytesize'] == 24
    # make sure git-annex recorded an unmodified URL
    assert any(testurl in r.get('urls', [])
               for r in ds.repo.whereis('dummy', output='full').values())


def test_uncurl(existing_dataset, tmp_path):
    archive_path = tmp_path
    create_tree(archive_path, {'lvlA1': {'lvlB2_flavor1.tar': 'data_A1B2F1'}})
    ds = existing_dataset
    archive_path = Path(archive_path)
    dsca = ds.repo.call_annex
    dsca(['initremote', 'myuncurl'] + std_initargs + [
        'match=bingofile://(?P<basepath>.*)/(?P<lvlA>[^/]+)/(?P<lvlB>[^/]+)_(?P<flavor>.*)$ someothermatch',
        'url=file://{basepath}/{lvlA}/{lvlB}_{flavor}',
    ])
    data_url = (archive_path / 'lvlA1' / 'lvlB2_flavor1.tar').as_uri()
    # prefix the URL so git-annex has no idea how to handle it
    # (same as migrating from an obsolete system with no support anymore)
    data_url = f'bingo{data_url}'
    dsca(['addurl', '--file', 'data_A1B2F1.dat', data_url])
    assert (ds.pathobj / 'data_A1B2F1.dat').read_text() == 'data_A1B2F1'
    # file is known to be here and a (uncurl) remote
    assert len(ds.repo.whereis('data_A1B2F1.dat')) == 2
    # must survive an fsck (CHECKPRESENT)
    dsca(['fsck', '-q', '-f', 'myuncurl'])


# RIA tooling is not working for this test on windows
# https://github.com/datalad/datalad/issues/7212
@skip_if_on_windows
def test_uncurl_ria_access(tmp_path, no_result_rendering):
    """
    - create dataset with test file and push into RIA store
    - create INDEPENDENT dataset and 'addurl' test file directly from RIA
    - test that addurls work without any config, just initremote with no
      custom settings
    - now move RIA and hence break URL
    - fix1: only using a URL template, point to dataset dir in RIA store
      plus some always available key-properties
    - alternative fix2: simpler template, plus match expression to
      "understand" some structural aspects of RIA and reuse them
    """
    # we create a dataset to bootstrap the test setup, with on file
    # of known content
    srcds = EnsureDataset()(tmp_path / 'srcds').ds.create()
    testfile_content = 'mikewashere!'
    (srcds.pathobj / 'testfile.txt').write_text(testfile_content)
    srcds.save()
    # pull out some essential properties for the underlying key for later
    # use in this test
    testkey_props = srcds.status(
        'testfile.txt', annex='basic', return_type='item-or-list')
    testkey_props = {
        k: v for k, v in testkey_props.items()
        if k in ('key', 'hashdirmixed', 'hashdirlower')
    }

    # establish a RIA sibling and push
    baseurl = (tmp_path / "ria").as_uri()
    srcds.create_sibling_ria(
        # use a ria+file:// URL for simplicity
        f'ria+{baseurl}',
        name='ria',
        new_store_ok=True,
    )
    srcds.push(to='ria')
    # setup is done

    # start of the actual test
    # create a fresh dataset
    ds = EnsureDataset()(tmp_path / 'testds').ds.create()
    dsca = ds.repo.call_annex
    # we add uncurl WITH NO config whatsoever.
    # this must be enough to be able to use the built-in downloaders
    target_fname = 'mydownload.txt'
    dsca(['initremote', 'myuncurl'] + std_initargs)
    dsca(['addurl', '--file', target_fname,
          # we download from the verbatim, hand-crafted URL
          f'{baseurl}/{srcds.id[:3]}/{srcds.id[3:]}/annex/objects/'
          f'{testkey_props["hashdirmixed"]}'
          f'{testkey_props["key"]}/{testkey_props["key"]}'
    ])
    assert (ds.pathobj / target_fname).read_text() == testfile_content
    # make sure the re-downloaded key ends up having the same keyname in
    # the new dataset
    assert ds.status(
        target_fname, annex='basic', return_type='item-or-list',
        )['key'] == testkey_props['key']

    # now we drop the key...
    ds.drop(target_fname)
    assert not ds.status(
        target_fname, annex='availability', return_type='item-or-list',
        )['has_content']
    # ...and we move the RIA store to break the recorded
    # URL (simulating an infrastructure change)
    (tmp_path / 'ria').rename(tmp_path / 'ria_moved')

    # verify that no residual magic makes data access possible
    with pytest.raises(IncompleteResultsError):
        ds.get(target_fname)

    # fix it via an access URL config,
    # point directly via a hard-coded dataset ID
    # NOTE: last line is no f-string!
    url_tmpl = (
        tmp_path / "ria_moved" / srcds.id[:3] / srcds.id[3:]
        ).as_uri() + '/annex/objects/{annex_dirhash}/{annex_key}/{annex_key}'
    ds.configuration(
        'set', f'remote.myuncurl.uncurl-url={url_tmpl}')
    # confirm checkpresent acknowledges this
    dsca(['fsck', '-q', '-f', 'myuncurl'])
    # confirm transfer_retrieve acknowledges this
    ds.get(target_fname)

    # but we can also do without hard-coding anything, so let's drop again
    ds.drop(target_fname)
    assert not ds.status(
        target_fname, annex='availability', return_type='item-or-list',
        )['has_content']

    # for that we need to add a match expression that can "understand"
    # the original URL. All we need is to distinguish the old base path
    # from the structured components in the RIA store (for the
    # latter we can simple account for 4 levels of sudirs
    ds.configuration(
        'set',
        'remote.myuncurl.uncurl-match='
        'file://(?P<basepath>.*)/(?P<dsdir>[^/]+/[^/]+)/annex/objects/.*$',
        scope='local')
    # NOTE: last line is no f-string!
    url_tmpl = (tmp_path / "ria_moved").as_uri() \
        + '/{dsdir}/annex/objects/{annex_dirhash}/{annex_key}/{annex_key}'
    ds.configuration(
        'set', f'remote.myuncurl.uncurl-url={url_tmpl}',
        scope='local')
    # confirm checkpresent acknowledges this
    dsca(['fsck', '-q', '-f', 'myuncurl'])
    # confirm transfer_retrieve acknowledges this
    ds.get(target_fname)
    assert (ds.pathobj / target_fname).read_text() == testfile_content


@skip_if_root  # see https://github.com/datalad/datalad-next/issues/525
def test_uncurl_store(tmp_path, existing_dataset, no_result_rendering):
    ds = existing_dataset
    testfile = ds.pathobj / 'testfile1.txt'
    testfile_content = 'uppytyup!'
    testfile.write_text(testfile_content)
    ds.save()
    dsca = ds.repo.call_annex
    # init the remote with a template that places keys in the same structure
    # as annex/objects within a bare remote repo
    dsca(['initremote', 'myuncurl'] + std_initargs + [
        # intentional double-braces at the end to get templates into the template
        f'url={(tmp_path).as_uri()}/{{annex_dirhash_lower}}{{annex_key}}/{{annex_key}}',
    ])
    # store file at remote
    dsca(['copy', '-t', 'myuncurl', str(testfile)])
    # let remote verify presence
    dsca(['fsck', '-q', '-f', 'myuncurl'])
    # doublecheck
    testfile_props = ds.status(testfile, annex='basic',
                               return_type='item-or-list')
    assert (tmp_path / testfile_props['hashdirlower'] /
            testfile_props['key'] / testfile_props['key']
        ).read_text() == testfile_content
    # we have no URLs recorded
    assert all(not v['urls']
               for v in ds.repo.whereis(str(testfile), output='full').values())
    # yet we can retrieve via uncurl, because local key properties are enough
    # to fill the template
    ds.drop(testfile)
    assert not ds.status(
        testfile, annex='availability', return_type='item-or-list',
        )['has_content']
    dsca(['copy', '-f', 'myuncurl', str(testfile)])
    assert testfile.read_text() == testfile_content

    if on_windows:
        # remaining bits assume POSIX FS
        return

    ds.config.set(
        'remote.myuncurl.uncurl-url',
        # same as above, but root with no write-permissions
        'file:///youshallnotpass/{annex_dirhash_lower}{annex_key}/{annex_key}',
        scope='local',
    )
    with pytest.raises(CommandError):
        dsca(['fsck', '-q', '-f', 'myuncurl'])
    with pytest.raises(CommandError):
        dsca(['copy', '-t', 'myuncurl', str(testfile)])


def test_uncurl_store_via_ssh(
        sshserver, existing_dataset, no_result_rendering):
    ds = existing_dataset
    testfile = ds.pathobj / 'testfile1.txt'
    testfile_content = 'uppytyup!'
    testfile.write_text(testfile_content)
    ds.save()
    dsca = ds.repo.call_annex
    # init the remote with a template that places keys in the same structure
    # as annex/objects within a bare remote repo
    dsca(['initremote', 'myuncurl'] + std_initargs + [
        # intentional double-braces at the end to get templates into the template
        f'url={sshserver[0]}/{{annex_key}}',
    ])
    # store file at remote
    dsca(['copy', '-t', 'myuncurl', str(testfile)])
    # let remote verify presence
    dsca(['fsck', '-q', '-f', 'myuncurl'])


def test_uncurl_remove(existing_dataset, tmp_path, no_result_rendering):
    testfile = tmp_path / 'testdeposit' / 'testfile1.txt'
    testfile_content = 'uppytyup!'
    testfile.parent.mkdir()
    testfile.write_text(testfile_content)
    ds = existing_dataset
    dsca = ds.repo.call_annex
    # init without URL template
    dsca(['initremote', 'myuncurl'] + std_initargs)
    # add the testdeposit by URL
    target_fname = ds.pathobj / 'target1.txt'
    dsca(['addurl', '--file', str(target_fname), testfile.as_uri()])
    # it will not drop without a URL tmpl
    # see test_uncurl_remove_no_tmpl() for rational
    with pytest.raises(CommandError):
        dsca(['drop', '-f', 'myuncurl', str(target_fname)])
    assert testfile.read_text() == testfile_content

    # now make it possible
    # use the simplest possible match expression
    ds.configuration(
        'set',
        'remote.myuncurl.uncurl-match=file://(?P<allofit>.*)$',
        scope='local')
    # and the presence of a tmpl enables deletion
    ds.configuration(
        'set', 'remote.myuncurl.uncurl-url=file://{allofit}',
        scope='local')
    dsca(['drop', '-f', 'myuncurl', str(target_fname)])
    assert not testfile.exists()


# >30s
def test_uncurl_testremote(tmp_path, existing_dataset):
    "Point git-annex's testremote at uncurl"
    ds = existing_dataset
    dsca = ds.repo.call_annex
    dsca(['initremote', 'myuncurl'] + std_initargs
         # file://<basepath>/key
         + [f'url=file://{tmp_path}/{{annex_key}}'])
    # Temporarily disable this until
    # https://github.com/datalad/datalad-dataverse/issues/127
    # is sorted out. Possibly via
    # https://git-annex.branchable.com/bugs/testremote_is_not_honoring_--backend
    if not on_windows:
        # not running with --fast to also cover key chunking
        dsca(['testremote', '--quiet', 'myuncurl'])



"""Assorted utility functions

.. currentmodule:: datalad_next.utils
.. autosummary::
   :toctree: generated

   DataladAuth
   MultiHash
   check_symlink_capability
   chpwd
   ensure_list
   external_versions
   log_progress
   parse_www_authenticate
   patched_env
   rmtree
   get_specialremote_param_dict
   get_specialremote_credential_properties
   update_specialremote_credential
   needs_specialremote_credential_envpatch
   get_specialremote_credential_envpatch
"""

from datalad.utils import (
    # TODO REMOVE FOR V2.0
    Path,
    check_symlink_capability,
    chpwd,
    # TODO REMOVE FOR V2.0
    ensure_bool,
    ensure_list,
    # TODO https://github.com/datalad/datalad-next/issues/626
    get_dataset_root,
    # TODO REMOVE FOR V2.0
    # only needed for `interface_utils` patch and better be imported
    # in there
    getargspec,
    # TODO REMOVE FOR V2.0
    get_wrapped_class,
    # TODO REMOVE FOR V2.0
    knows_annex,
    # TODO REMOVE FOR V2.0
    # in datalad_next.consts
    on_linux,
    # TODO REMOVE FOR V2.0
    # in datalad_next.consts
    on_windows,
    # TODO REMOVE FOR V2.0
    rmtemp,
    rmtree,
    # TODO REMOVE FOR V2.0
    # only a test utility and should move there
    swallow_outputs,
)
# TODO REMOVE FOR V2.0
# internal helper of create_sibling_webdav
from datalad.distribution.utils import _yield_ds_w_matching_siblings
from datalad.support.external_versions import external_versions

# TODO REMOVE FOR V2.0
from datalad_next.credman import CredentialManager
from .log import log_progress
from .multihash import MultiHash
from .requests_auth import (
    DataladAuth,
    parse_www_authenticate,
)
from .specialremote import (
    get_specialremote_param_dict,
    get_specialremote_credential_properties,
    update_specialremote_credential,
    specialremote_credential_envmap,
    needs_specialremote_credential_envpatch,
    get_specialremote_credential_envpatch,
)
from .patch import patched_env

# TODO REMOVE EVERYTHING BELOW FOR V2.0
# https://github.com/datalad/datalad-next/issues/611
from typing import (
    Any,
    Dict,
)


class ParamDictator:
    """Parameter dict access helper

    This class can be used to wrap a dict containing function parameter
    name-value mapping, and get/set values by parameter name attribute
    rather than via the ``__getitem__`` dict API.
    """
    def __init__(self, params: Dict):
        self.__params = params

    def __getattr__(self, attr: str):
        if attr.startswith('__'):
            return super().__getattribute__(attr)
        return self.__params[attr]

    def __setattr__(self, attr: str, value: Any):
        if attr.startswith('_ParamDictator__'):
            super().__setattr__(attr, value)
        else:
            self.__params[attr] = value



"""Post DataLad config overrides CLI/ENV as GIT_CONFIG items in process ENV

This enables their propagation to any subprocess. This includes the
specification of overrides via the ``datalad -c ...`` option of the
main CLI entrypoint.
"""

from datalad.config import _update_from_env as _update_from_datalad_env
from datalad.cli.helpers import _parse_overrides_from_cmdline

from datalad_next.config.utils import (
    get_gitconfig_items_from_env,
    set_gitconfig_items_in_env,
)

from . import apply_patch


def parse_overrides_from_cmdline(cmdlineargs):
    # read from cmdlineargs first to error on any syntax issues
    # before any other processing
    cli_overrides = _parse_overrides_from_cmdline(cmdlineargs)

    # reuse datalad-core implementation of datalad-specific ENV parsing
    # for config items
    overrides = {}
    _update_from_datalad_env(overrides)

    # let CLI settings override any ENV -- in-line with the behavior of Git
    overrides.update(cli_overrides)

    # read any existing GIT_CONFIG ENV vars and superimpose our
    # overrides on them, repost in ENV using git-native approach.
    # This will apply the overrides to any git(-config) calls
    # in this process and any subprocess
    gc_overrides = get_gitconfig_items_from_env()
    gc_overrides.update(overrides)
    set_gitconfig_items_in_env(gc_overrides)

    # we do not actually disclose any of these overrides.
    # the CLI runs a `datalad.cfg.reload(force=True)`
    # immediately after executing this function and thereby
    # pulls in the overrides we just posted into the ENV
    # here. This change reduced the scope of
    # `datalad.cfg.overrides` to be mere instance overrides
    # and no longer process overrides. This rectifies the mismatch
    # between appearance and actual impact of this information
    # in the ConfigManager
    return {}


apply_patch(
    'datalad.cli.helpers', None, '_parse_overrides_from_cmdline',
    parse_overrides_from_cmdline,
    msg='Enable posting DataLad config overrides CLI/ENV as '
    'GIT_CONFIG items in process ENV',
)



"""Compute more than one hash for the same data in one go"""

from __future__ import annotations

import hashlib
from typing import (
    ByteString,
    Dict,
)


class NoOpHash:
    """Companion of :class:`MultiHash` that computes no hash at all

    This can be used wherever ``MultiHash`` would be used, because
    it implements its API. However, no hash is computed and no
    hexdigest is reported.
    """
    def __init__(self, algorithms: None = None):
        pass

    def update(self, data):
        pass

    def get_hexdigest(self):
        return {}


class MultiHash:
    """Compute any number of hashes as if computing just one

    Supports any hash algorithm supported by the ``hashlib`` module
    of the standard library.
    """
    def __init__(self, algorithms: list[str]):
        """
        Parameters
        ----------
        algorithms: list
          Hash names, must match the name of the algorithms in the
          ``hashlib`` module (case insensitive).
        """
        # yes, this will crash, if an invalid hash algorithm name
        # is given
        _hasher = []
        for h in algorithms:
            hr = getattr(hashlib, h.lower(), None)
            if hr is None:
                raise ValueError(f'unsupported hash algorithm {h}')
            _hasher.append(hr())
        self._hasher = dict(zip(algorithms, _hasher))

    def update(self, data: ByteString) -> None:
        """Updates all configured digests"""
        for h in self._hasher.values():
            h.update(data)

    def get_hexdigest(self) -> Dict[str, str]:
        """Returns a mapping of algorithm name to hexdigest for all algorithms
        """
        return {a: h.hexdigest() for a, h in self._hasher.items()}



"""Credential support for ``AnnexRepo.enable_remote()`` and ``siblings enable``

Supported targets for automatic credential deployments are determined
by ``needs_specialremote_credential_envpatch()``. At the time of this
writing this includes the git-annex built-in remote types ``webdav``,
``s3``, and ``glacier``.

This patch also changes the function to raise its custom exception
with the context of an original underlying exception for better
error reporting.
"""
import logging
import os
import re
from unittest.mock import patch

from datalad.support.exceptions import (
    AccessDeniedError,
    AccessFailedError,
)
from datalad_next.exceptions import (
    CommandError,
)
from datalad_next.utils import (
    CredentialManager,
    ensure_list,
    get_specialremote_credential_envpatch,
    get_specialremote_param_dict,
    get_specialremote_credential_properties,
    needs_specialremote_credential_envpatch,
)
from . import apply_patch


# reuse logger from -core, despite the unconventional name
lgr = logging.getLogger('datalad.annex')


# This function is taken from datalad-core@2ed709613ecde8218a215dcb7d74b4a352825685
# datalad/support/annexrepo.py:AnnexRepo
def annexRepo__enable_remote(self, name, options=None, env=None):
    """Enables use of an existing special remote

    Parameters
    ----------
    name: str
        name, the special remote was created with
    options: list, optional
    """
    # MIH thinks there should be no `env` argument at all
    # https://github.com/datalad/datalad/issues/5162
    # if it would not be there, this whole dance is pretty much
    # obsolete
    env = env or self._git_runner.env

    # an enableremote can do pretty much anything, including a type change.
    # in order to be able to determine whether credentials *will* be needed,
    # we have to look ahead and form the special remote parameters that will
    # be there at the end -- more or less

    # pull info for present config
    sp_remotes = {v['name']: dict(v, uuid=k) for k, v in self.get_special_remotes().items()}
    remote_info = sp_remotes.get(name, {})
    # TODO if remote_info is empty, we can fail right here
    if options:
        # and now update with given params
        remote_info.update(get_specialremote_param_dict(options))

    # careful here, `siblings()` also calls this for regular remotes, check
    # for a known type
    if 'type' in remote_info \
            and needs_specialremote_credential_envpatch(remote_info['type']):
        # see if we can identify any matching credentials
        credprops = get_specialremote_credential_properties(remote_info)
        credman = None
        credspec = None
        if credprops:
            credman = CredentialManager(self.config)
            creds = credman.query(_sortby='last-used', **credprops)
            if creds:
                # found one
                credspec = creds[0]
        # TODO manual entry could be supported here too! (also see at the end)
        if env:
            env.copy()

        if credspec:
            credpatch = get_specialremote_credential_envpatch(
                remote_info['type'], credspec[1])
            if credpatch:
                if not env:
                    env = os.environ.copy()
                env.update(credpatch)

    try:
        with patch.object(self._git_runner, 'env', env):
            # TODO: outputs are nohow used/displayed. Eventually convert to
            # to a generator style yielding our "dict records"
            self.call_annex(['enableremote', name] + ensure_list(options))
    except CommandError as e:
        if re.match(r'.*StatusCodeException.*statusCode = 401', e.stderr):
            raise AccessDeniedError(e.stderr) from e
        elif 'FailedConnectionException' in e.stderr:
            raise AccessFailedError(e.stderr) from e
        else:
            raise e
    self.config.reload()
    # TODO when manual credential entry is supported,
    # implement store-after-success here


apply_patch(
    'datalad.support.annexrepo', 'AnnexRepo', 'enable_remote',
    annexRepo__enable_remote,
    msg='Apply datalad-next patch to annexrepo.py:AnnexRepo.enable_remote')



"""Improve ``CommandError`` rendering and add ``returncode`` alias for ``code``


This patch does two things:

It overwrites ``__repr__``, otherwise ``CommandError` would use
``RuntimeError``'s variant and ignore all additional structured information
except for ``.msg`` -- which is frequently empty and confuses with a
`CommandError('')` display.

It adds a ``returncode`` alias for ``code``. This unifies return code access
between ``CommandError`` and `Popen``-like objects, which usually have a
``returncode`` attribute.
"""

from datalad.runner.exception import CommandError


def commanderror_repr(self) -> str:
    return self.to_str()


CommandError.__repr__ = commanderror_repr


# Basic alias idea taken from here:
# <https://stackoverflow.com/questions/4017572/how-can-i-make-an-alias-to-a-non-function-member-attribute-in-a-python-class>
_commanderror_aliases = {
    'returncode': 'code',
}

def commanderror_getattr(self, item):
    return object.__getattribute__(self, _commanderror_aliases.get(item, item))

def commanderror_setattr(self, key, value):
    if key == '_aliases':
        raise AttributeError('Cannot set `_aliases`')
    return object.__setattr__(self, _commanderror_aliases.get(key, key), value)


CommandError.__getattr__ = commanderror_getattr
CommandError.__setattr__ = commanderror_setattr



from __future__ import annotations

import os
from pathlib import Path
import subprocess

from datalad_next.exceptions import CapturedException

from .iter_subproc import (
    CommandError,
    iter_subproc,
)


def _call_git(
    args: list[str],
    *,
    capture_output: bool = False,
    cwd: Path | None = None,
    check: bool = False,
    text: bool | None = None,
    input: str | bytes | None = None,
    force_c_locale: bool = False,
) -> subprocess.CompletedProcess:
    """Wrapper around ``subprocess.run`` for calling Git command

    ``args`` is a list of argument for the Git command. This list must not
    contain the Git executable itself. It will be prepended (unconditionally)
    to the arguments before passing them on.

    If ``force_c_locale`` is ``True`` the environment of the Git process
    is altered to ensure output according to the C locale. This is useful
    when output has to be processed in a locale invariant fashion.

    All other argument are pass on to ``subprocess.run()`` verbatim.
    """
    env = None
    if force_c_locale:
        env = dict(os.environ, LC_ALL='C')

    # make configurable
    git_executable = 'git'
    cmd = [git_executable, *args]
    try:
        return subprocess.run(
            cmd,
            capture_output=capture_output,
            cwd=cwd,
            check=check,
            text=text,
            input=input,
            env=env,
        )
    except subprocess.CalledProcessError as e:
        # TODO we could support post-error forensics, but some client
        # might call this knowing that it could fail, and may not
        # appreciate the slow-down. Add option `expect_fail=False`?
        #
        # normalize exception to datalad-wide standard
        raise CommandError(
            cmd=cmd,
            code=e.returncode,
            stdout=e.stdout,
            stderr=e.stderr,
            cwd=cwd,
        ) from e


def call_git(
    args: list[str],
    *,
    cwd: Path | None = None,
    force_c_locale: bool = False,
) -> None:
    """Call Git with no output capture, raises on non-zero exit.

    If ``cwd`` is not None, the function changes the working directory to
    ``cwd`` before executing the command.

    If ``force_c_locale`` is ``True`` the environment of the Git process
    is altered to ensure output according to the C locale. This is useful
    when output has to be processed in a locale invariant fashion.
    """
    _call_git(
        args,
        capture_output=False,
        cwd=cwd,
        check=True,
        force_c_locale=force_c_locale,
    )


def call_git_success(
    args: list[str],
    *,
    cwd: Path | None = None,
    capture_output: bool = False,
) -> bool:
    """Call Git and report success or failure of the command

    ``args`` is a list of arguments for the Git command. This list must not
    contain the Git executable itself. It will be prepended (unconditionally)
    to the arguments before passing them on.

    If ``cwd`` is not None, the function changes the working directory to
    ``cwd`` before executing the command.

    If ``capture_output`` is ``True``, process output is captured, but not
    returned. By default process output is not captured.
    """
    try:
        _call_git(
            args,
            capture_output=capture_output,
            cwd=cwd,
            check=True,
        )
    except CommandError as e:
        CapturedException(e)
        return False
    return True


def call_git_lines(
    args: list[str],
    *,
    cwd: Path | None = None,
    input: str | None = None,
    force_c_locale: bool = False,
) -> list[str]:
    """Call Git for any (small) number of lines of output

    ``args`` is a list of arguments for the Git command. This list must not
    contain the Git executable itself. It will be prepended (unconditionally)
    to the arguments before passing them on.

    If ``cwd`` is not None, the function changes the working directory to
    ``cwd`` before executing the command.

    If ``input`` is not None, the argument becomes the subprocess’s stdin.
    This is intended for small-scale inputs. For call that require processing
    large inputs, ``iter_git_subproc()`` is to be preferred.

    If ``force_c_locale`` is ``True`` the environment of the Git process
    is altered to ensure output according to the C locale. This is useful
    when output has to be processed in a locale invariant fashion.

    Raises
    ------
    CommandError if the call exits with a non-zero status.
    """
    res = _call_git(
        args,
        capture_output=True,
        cwd=cwd,
        check=True,
        text=True,
        input=input,
        force_c_locale=force_c_locale,
    )
    return res.stdout.splitlines()


def call_git_oneline(
    args: list[str],
    *,
    cwd: Path | None = None,
    input: str | None = None,
    force_c_locale: bool = False,
) -> str:
    """Call Git for a single line of output

    If ``cwd`` is not None, the function changes the working directory to
    ``cwd`` before executing the command.

    If ``input`` is not None, the argument becomes the subprocess’s stdin.
    This is intended for small-scale inputs. For call that require processing
    large inputs, ``iter_git_subproc()`` is to be preferred.

    If ``force_c_locale`` is ``True`` the environment of the Git process
    is altered to ensure output according to the C locale. This is useful
    when output has to be processed in a locale invariant fashion.

    Raises
    ------
    CommandError if the call exits with a non-zero status.
    AssertionError if there is more than one line of output.
    """
    lines = call_git_lines(args, cwd=cwd, input=input,
                           force_c_locale=force_c_locale)
    if len(lines) > 1:
        raise AssertionError(
            f"Expected Git {args} to return a single line, but got {lines}"
        )
    return lines[0]


def iter_git_subproc(
    args: list[str],
    **kwargs
):
    """``iter_subproc()`` wrapper for calling Git commands

    All argument semantics are identical to those of ``iter_subproc()``,
    except that ``args`` must not contain the Git binary, but need to be
    exclusively arguments to it. The respective `git` command/binary is
    automatically added internally.
    """
    cmd = ['git']
    cmd.extend(args)

    return iter_subproc(cmd, **kwargs)



"""Essential tooling for implementing DataLad commands

This module provides the advanced command base class
:class:`ValidatedInterface`, for implementing commands with uniform argument
validation and structured error reporting.

Beyond that, any further components necessary to implement command are imported
in this module to offer a one-stop-shop experience. This includes
``build_doc``, ``datasetmethod``, and ``eval_results``, among others.

.. currentmodule:: datalad_next.commands
.. autosummary::
   :toctree: generated

   CommandResult
   CommandResultStatus
   status.StatusResult
"""
from __future__ import annotations

from typing import Dict

from datalad.interface.base import (
    Interface,
    build_doc,
)
from datalad.interface.results import get_status_dict
from datalad.interface.utils import generic_result_renderer
from datalad.interface.base import eval_results
from datalad.support.param import Parameter

from datalad_next.constraints import (
    EnsureCommandParameterization,
    ParameterConstraintContext,
)
from datalad_next.datasets import datasetmethod
from .results import (
    CommandResult,
    CommandResultStatus,
)

class ValidatedInterface(Interface):
    """Alternative base class for commands with uniform parameter validation

    .. note::
       This interface is a draft. Usage is encouraged, but future changes
       are to be expected.

    Commands derived from the traditional ``Interface`` class have no built-in
    input parameter validation beyond CLI input validation of individual
    parameters. Consequently, each command must perform custom parameter
    validation, which often leads to complex boilerplate code that is largely
    unrelated to the purpose of a particular command.

    This class is part of a framework for uniform parameter validation,
    regardless of the target API (Python, CLI, GUI). The implementation of
    a command's ``__call__`` method can focus on the core purpose of the
    command, while validation and error handling can be delegated elsewhere.

    A validator for all individual parameters and the joint-set of all
    parameters can be provided through the :meth:`get_parameter_validator`
    method.

    To transition a command from ``Interface`` to ``ValidatedInterface``,
    replace the base class declaration and declare a ``_validator_`` class
    member. Any ``constraints=`` declaration for ``Parameter`` instances
    should either be removed, or moved to the corresponding entry in
    ``_validator_``.
    """
    _validator_: EnsureCommandParameterization | None = None

    @classmethod
    def get_parameter_validator(cls) -> EnsureCommandParameterization | None:
        """Returns a validator for the entire parameter set of a command

        If parameter validation shall be performed, this method must return an
        instance of
        :class:`~datalad_next.constraints.EnsureCommandParameterization`.
        All parameters will be passed through this validator, and only the its
        output will be passed to the underlying command's ``__call__`` method.

        Consequently, the core implementation of a command only needs to
        support the output values of the validators declared by itself.

        Factoring out input validation, normalization, type coercion etc. into
        a dedicated component also makes it accessible for upfront validation
        and improved error reporting via the different DataLad APIs.

        If a command does not implement parameter validation in this fashion,
        this method must return ``None``.

        The default implementation returns the ``_validator_`` class member.
        """
        return cls._validator_



from __future__ import annotations

from datasalad.runners import iter_subproc
import os
import sys
from pathlib import PurePosixPath
from shlex import quote as posix_quote

import pytest
from more_itertools import consume

import datalad
from datalad.tests.utils_pytest import (
    on_windows,
    skip_if,
)
from datalad_next.exceptions import CommandError
from datalad_next.url_operations.ssh import ssh_url2openargs
from ..response_generators import (
    FixedLengthResponseGeneratorPosix,
    FixedLengthResponseGeneratorPowerShell,
    VariableLengthResponseGeneratorPosix,
    VariableLengthResponseGeneratorPowerShell,
    lgr as response_generator_lgr
)
from ..shell import (
    ShellCommandExecutor,
    shell,
)
from .. import posix


# Some files that are usually found on POSIX systems, i.e. Linux, OSX
common_files = [b'/etc/passwd', b'/etc/shells']

# Select "challenging" file names that need proper quoting based on Windows, on
# POSIX, and on FAT file systems.
if os.getenv('TMPDIR', '').startswith('/crippledfs'):
    upload_file_name = "up 1"
    download_file_name = "down 1"
    files_to_delete = ('f 1', 'f 2', 'f 3')
elif on_windows:
    upload_file_name = "upload 123"
    download_file_name = "download 123"
    files_to_delete = ('f 1', 'f 2', 'f 3')
else:
    upload_file_name = "upload $123 \"'"
    download_file_name = "download $123 \" ' "
    files_to_delete = ('f $1', 'f \\2 " ', 'f 3 \' ')


def _get_cmdline(ssh_url: str):
    args, parsed = ssh_url2openargs(ssh_url, datalad.cfg)
    return ['ssh'] + args, parsed.path


@pytest.mark.parametrize('file_name', common_files)
def test_basic_functionality(sshserver, file_name):
    ssh_url = sshserver[0]
    with shell(_get_cmdline(ssh_url)[0]) as ssh:
        _check_ls_result(ssh, file_name)


def test_basic_functionality_multi(sshserver):
    # Similar to `test_basic_functionality`, but executes all commands on the
    # same connection.
    ssh_url = sshserver[0]
    with shell(_get_cmdline(ssh_url)[0]) as ssh_executor:
        for file_name in common_files:
            _check_ls_result(ssh_executor, file_name)


def _quote_file_name(file_name: bytes, *, encoding: str = 'utf-8') -> bytes:
    return posix_quote(file_name.decode(encoding)).encode(encoding)


def _check_ls_result(ssh_executor, file_name: bytes):
    quoted_file_name = _quote_file_name(file_name)
    result = ssh_executor(b'ls ' + quoted_file_name)
    assert result.stdout == file_name + b'\n'
    result = ssh_executor('ls ' + quoted_file_name.decode())
    assert result.stdout == file_name + b'\n'


def test_return_code_functionality(sshserver):
    ssh_url = sshserver[0]
    with shell(_get_cmdline(ssh_url)[0]) as ssh:
        result = ssh(b'bash -c "exit 123"')
        assert result.returncode == 123


@pytest.mark.parametrize('cmd,expected', [
    (b'echo 0123456789', b'0123456789\n'),
    (b'echo -n 0123456789', b'0123456789')
])
def test_stdout_forwarding(sshserver, cmd, expected):
    ssh_url = sshserver[0]
    with shell(_get_cmdline(ssh_url)[0]) as ssh:
        _check_echo_result(ssh, cmd, expected)


def test_stdout_forwarding_multi(sshserver):
    # Similar to `test_stdout_forwarding`, but executes all commands on the
    # same connection.
    ssh_url = sshserver[0]
    with shell(_get_cmdline(ssh_url)[0]) as ssh:
        for cmd, expected in [(b'echo 0123456789', b'0123456789\n'),
                              (b'echo -n 0123456789', b'0123456789')]:
            _check_echo_result(ssh, cmd, expected)


def _check_echo_result(ssh: ShellCommandExecutor, cmd: bytes, expected: bytes):
    result = ssh(cmd)
    assert result.stdout == expected
    assert result.returncode == 0


def test_exit_if_unlimited_stdin_is_closed(sshserver):
    # Check that the test terminates if stdin is closed

    ssh_url, local_path = sshserver
    ssh_args, ssh_path = _get_cmdline(ssh_url)
    test_file_name = 'cat-123'
    # We know the ssh-server is on a POSIX system
    ssh_path = (ssh_path + '/' + test_file_name).encode()
    with \
            shell(ssh_args) as ssh_executor, \
            iter_subproc([sys.executable, '-c', 'print("0123456789")']) as cat_feed:

        response_generator = ssh_executor.start(b'cat >' + ssh_path, stdin=cat_feed)
        ssh_executor.close()
        consume(response_generator)
        assert response_generator.returncode == 0
        assert (local_path / test_file_name).read_text() == '0123456789\n'


def test_continuation_after_stdin_reading(sshserver):
    # check that the connection continues to work, after stdin was fed into the
    # remote command.

    ssh_url, local_path = sshserver
    ssh_args, ssh_path = _get_cmdline(ssh_url)
    feed_command = [sys.executable, '-c', 'print("0123456789", end="")']
    with \
            shell(ssh_args) as ssh_executor, \
            iter_subproc(feed_command) as dd_feed_1, \
            iter_subproc(feed_command) as dd_feed_2:

        for file_name, feed in (('dd-123', dd_feed_1), ('dd-456', dd_feed_2)):
            server_path = (ssh_path + '/' + file_name).encode()
            result = ssh_executor(
                b'dd bs=1 count=10 of=' + server_path,
                stdin=feed
            )
            assert result.returncode == 0
            assert (local_path / file_name).read_text() == '0123456789'

        _check_ls_result(ssh_executor, common_files[0])


def test_upload(sshserver, tmp_path):
    ssh_url, local_path = sshserver
    ssh_args, ssh_path = _get_cmdline(ssh_url)
    content = '0123456789'
    upload_file = tmp_path / upload_file_name
    upload_file.write_text(content)
    progress = []
    with shell(ssh_args) as ssh_executor:
        # perform an operation on the remote shell
        _check_ls_result(ssh_executor, common_files[0])

        # upload file to server and verify its content
        result = posix.upload(
            ssh_executor,
            upload_file,
            PurePosixPath(ssh_path + '/' + upload_file_name),
            progress_callback=lambda a, b: progress.append((a, b))
        )
        assert result.returncode == 0
        assert (local_path / upload_file_name).read_text() == content
        assert len(progress) > 0

        # perform another operation on the remote shell to ensure functionality
        _check_ls_result(ssh_executor, common_files[0])


def test_download_ssh(sshserver, tmp_path):
    ssh_url, local_path = sshserver
    ssh_args, ssh_path = _get_cmdline(ssh_url)
    content = '0123456789'
    server_file = local_path / download_file_name
    server_file.write_text(content)
    download_file = tmp_path / download_file_name
    progress = []
    with shell(ssh_args) as ssh_executor:
        # perform an operation on the remote shell
        _check_ls_result(ssh_executor, common_files[0])

        # download file from server and verify its content
        result = posix.download(
            ssh_executor,
            PurePosixPath(ssh_path + '/' + download_file_name),
            download_file,
            progress_callback=lambda a, b: progress.append((a, b))
        )
        assert result.returncode == 0
        assert download_file.read_text() == content
        assert len(progress) > 0

        # perform another operation on the remote shell to ensure functionality
        _check_ls_result(ssh_executor, common_files[0])


# This test only works on Posix-like systems because it executes a local
# bash command.
@skip_if(on_windows)
def test_download_local_bash(tmp_path):
    content = '0123456789'
    download_file = tmp_path / download_file_name
    download_file.write_text(content)
    result_file = tmp_path / ('result' + download_file_name)
    progress = []
    with shell(['bash']) as bash:
        _check_ls_result(bash, common_files[0])

        # download file from server and verify its content
        posix.download(
            bash,
            PurePosixPath(download_file),
            result_file,
            progress_callback=lambda a, b: progress.append((a, b)),
        )
        assert result_file.read_text() == content
        assert len(progress) > 0

        # perform another operation on the remote shell to ensure functionality
        _check_ls_result(bash, common_files[0])


# This test only works on Posix-like systems because it executes a local bash
@skip_if(on_windows)
def test_upload_local_bash(tmp_path):
    content = '0123456789'
    upload_file = tmp_path / upload_file_name
    upload_file.write_text(content)
    result_file = tmp_path / ('result' + upload_file_name)
    progress = []
    with shell(['bash']) as bash:
        _check_ls_result(bash, common_files[0])

        # upload file to server and verify its content
        posix.upload(
            bash,
            upload_file,
            PurePosixPath(result_file),
            progress_callback=lambda a, b: progress.append((a, b)),
        )
        assert result_file.read_text() == content
        assert len(progress) > 0

        # perform another operation on the remote shell to ensure functionality
        _check_ls_result(bash, common_files[0])


# This test only works on Posix-like systems because it executes a local bash
@skip_if(on_windows)
def test_upload_local_bash_error(tmp_path):
    content = '0123456789'
    source_file = tmp_path / 'upload_123'
    source_file.write_text(content)
    destination_file = PurePosixPath('/result_123')
    progress = []
    with shell(['bash']) as bash:
        _check_ls_result(bash, common_files[0])

        # upload file to a root on the server
        result = posix.upload(
            bash,
            source_file,
            destination_file,
            progress_callback=lambda a, b: progress.append((a, b)),
        )
        assert result.returncode != 0
        assert len(progress) > 0

        # perform another operation on the remote shell to ensure functionality
        _check_ls_result(bash, common_files[0])

        with pytest.raises(CommandError):
            posix.upload(bash, source_file, destination_file, check=True)

        # perform another operation on the remote shell to ensure functionality
        _check_ls_result(bash, common_files[0])


def test_delete(sshserver):
    ssh_url, local_path = sshserver
    ssh_args, ssh_path = _get_cmdline(ssh_url)

    with shell(ssh_args) as ssh_executor:
        for file in files_to_delete:
            (local_path / file).write_text(f'content_{file}')
            # verify that the remote files exist on the server
            _check_ls_result(ssh_executor, (ssh_path + '/' + file).encode())

        # delete files on server
        posix.delete(
            ssh_executor,
            [PurePosixPath(ssh_path) / file for file in files_to_delete],
            force=False,
        )

        # verify that the remote files were deleted
        for file in files_to_delete:
            assert not (local_path / file).exists()


def test_delete_error(sshserver):
    ssh_url, local_path = sshserver
    ssh_args, ssh_path = _get_cmdline(ssh_url)

    with shell(ssh_args) as ssh_executor:

        _check_ls_result(ssh_executor, common_files[0])

        # Try to delete a non-existing file
        result = posix.delete(
            ssh_executor,
            [PurePosixPath('/no-such-file')],
            force=False,
        )
        assert result.returncode != 0

        _check_ls_result(ssh_executor, common_files[0])

        with pytest.raises(CommandError):
            posix.delete(
                ssh_executor,
                [PurePosixPath('/no-such-file')],
                force=False,
                check=True,
            )
        _check_ls_result(ssh_executor, common_files[0])

        # Try to delete an existing file for which we are not authorized
        # Try to delete a non-existing file
        result = posix.delete(
            ssh_executor,
            [PurePosixPath('/etc/passwd')],
            force=False,
        )
        assert result.returncode != 0
        _check_ls_result(ssh_executor, common_files[0])

        with pytest.raises(CommandError):
            posix.delete(
                ssh_executor,
                [PurePosixPath('/etc/passwd')],
                force=False,
                check=True,
            )
        _check_ls_result(ssh_executor, common_files[0])


def test_returncode():
    with pytest.raises(RuntimeError):
        with shell(['ssh', 'xyz@localhost:22']):
            pass


@skip_if(not on_windows)
def test_powershell_basic():
    with shell(
            ['powershell', '-Command', '-'],
            zero_command_rg_class=VariableLengthResponseGeneratorPowerShell,
    ) as pwsh:
        r = pwsh(b'Get-ChildItem')
        assert r.returncode == 0
        r = pwsh(b'Get-ChildItem -Path C:\\')
        assert r.returncode == 0
        pwsh.close()


@skip_if(not on_windows)
def test_powershell_repr():
    with shell(
            ['powershell', '-Command', '-'],
            zero_command_rg_class=VariableLengthResponseGeneratorPowerShell,
    ) as pwsh:
        assert "ShellCommandExecutor(['powershell', '-Command', '-'])" == repr(pwsh)
        pwsh.close()


@skip_if(on_windows)
def test_posix_repr():
    with shell(['bash']) as ssh:
        assert "ShellCommandExecutor(['bash'])" == repr(ssh)
        ssh.close()


# This test only works on Posix-like systems because it executes a local
# bash command
@skip_if(on_windows)
def test_variable_length_reuse(monkeypatch):
    # This test ensures that the `VariableLengthResponseGenerator` can be
    # reused, e.g. after it was used for command zero, even if there is
    # unexpected output after the return code.
    def mocked_get_final_command(command: bytes) -> bytes:
        return (
            command + b' ; x=$?; echo -e -n "'
            + response_generator.end_marker
            + b'\\n"; echo -e "$x\\nsome stuff"\n'
        )

    log_messages = []

    def mocked_log(*args):
        log_messages.append(args[0])

    with shell(['bash']) as bash:
        response_generator = VariableLengthResponseGeneratorPosix(bash.stdout)
        monkeypatch.setattr(
            response_generator,
            'get_final_command',
            mocked_get_final_command
        )
        monkeypatch.setattr(response_generator_lgr, 'warning', mocked_log)
        result = bash(
            b'echo hello',
            response_generator=response_generator
        )
        assert result.stdout == b'hello\n'
        assert all(
            msg.startswith('unexpected output after return code: ')
            for msg in log_messages
        )
        bash.close()


# This test only works on Posix-like systems because it executes a local bash
@skip_if(on_windows)
def test_bad_zero_command(monkeypatch):
    monkeypatch.setattr(
        VariableLengthResponseGeneratorPosix,
        'zero_command',
        b'tessdsdsdt 0 -eq 1'
    )
    with pytest.raises(RuntimeError):
        with shell(['bash']):
            pass


# This test only works on Unix-like systems because it executes a local bash.
@skip_if(on_windows)
def test_fixed_length_response_generator_bash():
    with shell(['bash']) as bash:
        response_generator = FixedLengthResponseGeneratorPosix(
            bash.stdout,
            length=10
        )
        result = bash(
            b'echo -n 0123456789',
            response_generator=response_generator
        )
        assert result.stdout == b'0123456789'

        # Check that only 10 bytes are consumed and any excess bytes show up
        # in the return code.
        with pytest.raises(ValueError):
            result = bash(
                b'echo -n 0123456789abc',
                response_generator=response_generator
            )


# This test mostly works on Windows systems because it executes a local powershell.
@skip_if(not on_windows)
def test_fixed_length_response_generator_powershell():
    with shell(
            ['powershell', '-Command', '-'],
            zero_command_rg_class=VariableLengthResponseGeneratorPowerShell,
    ) as powershell:

        result = powershell(b'Write-Host -NoNewline 0123456789')
        assert result.returncode == 0
        if result.stdout.startswith(b'\nOops,'):
            pytest.skip(f'skipping test because powershell detected a bug')
        assert result.stdout == b'0123456789'

        # Check that only 10 bytes are consumed and any excess bytes show up
        # in the return code. Because the extra bytes are `'abc'`, the return
        # code is invalid, which leads to a `ValueError`.
        response_generator = FixedLengthResponseGeneratorPowerShell(
            powershell.stdout,
            length=10
        )
        with pytest.raises(ValueError):
            powershell(
                b'Write-Host -NoNewline 0123456789abc',
                response_generator=response_generator
            )


# This test only works on Unix-like systems because it executes a local bash.
@skip_if(on_windows)
def test_download_length_error():
    with shell(['bash']) as bash:
        response_generator = posix.DownloadResponseGeneratorPosix(bash.stdout)
        result = bash(b'unknown_file', response_generator=response_generator)
        assert result.stdout == b''
        assert result.returncode == 23

        # perform another operation on the remote shell to ensure functionality
        _check_ls_result(bash, common_files[0])


# This test does not work on Windows systems because it executes a local bash.
@skip_if(on_windows)
def test_download_error(tmp_path):
    progress = []
    with shell(['bash']) as bash:
        with pytest.raises(CommandError):
            posix.download(
                bash,
                PurePosixPath('/thisdoesnotexist'),
                tmp_path / 'downloaded_file',
                progress_callback=lambda a, b: progress.append((a, b)),
                check=True,
            )
        _check_ls_result(bash, common_files[0])

        result = posix.download(
            bash,
            PurePosixPath('/thisdoesnotexist'),
            tmp_path / 'downloaded_file',
            progress_callback=lambda a, b: progress.append((a, b)),
            check=False,
        )
        assert result.returncode not in (0, None)
        _check_ls_result(bash, common_files[0])



"""git-annex external backend XDLRA for git-remote-datalad-annex"""

from pathlib import Path
import zipfile

from .base import (
    Backend,
    BackendError,
    Master,
)


class DataladRepoAnnexBackend(Backend):
    """Implementation of an external git-annex backend

    This backend is tightly coupled to the `git-remote-datalad-annex`
    and hardly of any general utility. It is essentially aiming to be
    the leanest possible implementation to get git-annex to transport
    the content of two distinct files to and from a special remote.
    This backend is unlike most backends, because there is no fixed
    association of a particular file content to a particular key.
    In other words, the key content is expected to change without
    any change in the key name.

    Only two keys are supported:

    - ``XDLRA--refs``
    - ``XDLRA--repo-export``

    ``XDLRA--refs`` contains a "refs" list of a Git repository, similar
    to the output of ``git for-each-ref``. ``XDLRA--repo-export`` hold
    a ZIP archive of a bare Git repository.

    """
    def can_verify(self):
        # we can verify that a key matches the type of content
        # this is basically no more than a sanity check that a
        # download yielded something that we can work with for
        # downstream clone processing
        return True

    def is_stable(self):
        # the content behind a key is not always the same
        # in fact, it is typically different each time
        return False

    def is_cryptographically_secure(self):
        # we are not using any hashes
        return False

    def gen_key(self, local_file):
        localfile = Path(local_file)

        if _is_component_repoexport(localfile):
            return "XDLRA--repo-export"
        elif _is_component_refs(localfile):
            return "XDLRA--refs"
        else:
            # local_file is a TMP location, no use in reporting it
            raise BackendError('Unrecognized repository clone component')

    def verify_content(self, key, content_file):
        return self.gen_key(content_file) == key


def _is_component_refs(path):
    return path.read_text().endswith(' HEAD\n')


def _is_component_repoexport(path):
    return zipfile.is_zipfile(path)


def main():
    """Entry point for the backend utility"""
    master = Master()
    backend = DataladRepoAnnexBackend(master)
    master.LinkBackend(backend)
    master.Listen()



"""Exceptions to be used by all handlers"""

from __future__ import annotations

from typing import (
    Any,
)


class UrlOperationsRemoteError(Exception):
    def __init__(self, url, message=None, status_code: Any = None):
        # use base exception feature to store all arguments in a tuple
        # and have named properties to access them
        super().__init__(
            url,
            message,
            status_code,
        )

    def __str__(self):
        url, message, status_code = self.args
        if message:
            return message

        if status_code:
            return f"error {status_code} for {url!r}"

        return f"{self.__class__.__name__} for {url!r}"

    def __repr__(self) -> str:
        url, message, status_code = self.args
        return f"{self.__class__.__name__}(" \
               f"{url!r}, {message!r}, {status_code!r})"

    @property
    def url(self):
        return self.args[0]

    @property
    def message(self):
        return self.args[1]

    @property
    def status_code(self):
        return self.args[2]


class UrlOperationsResourceUnknown(UrlOperationsRemoteError):
    """A connection request succeeded in principle, but target was not found

    Equivalent of an HTTP404 response.
    """
    pass


class UrlOperationsInteractionError(UrlOperationsRemoteError):
    pass


class UrlOperationsAuthenticationError(UrlOperationsInteractionError):
    def __init__(self,
                 url: str,
                 credential: dict | None = None,
                 message: str | None = None,
                 status_code: Any = None):
        super().__init__(url, message=message, status_code=status_code)
        self.credential = credential


class UrlOperationsAuthorizationError(UrlOperationsRemoteError):
    def __init__(self,
                 url: str,
                 credential: dict | None = None,
                 message: str | None = None,
                 status_code: Any | None = None):
        super().__init__(url, message=message, status_code=status_code)
        self.credential = credential



import pytest

from datalad_next.exceptions import CommandError


def test_repr_str():
    # standard case of a command that failed with non-zero exit
    # many git/git-annex plumbing commands purposefully signal
    # statuses like this
    e = CommandError('some command', code=1)
    assert 'some command' in str(e)
    assert 'some command' in repr(e)

def test_returncode_code_alias():
    # check that `returncode` is an alias for `code`
    e = CommandError('some command', code=1)
    assert e.returncode == 1
    e.returncode = 2
    assert e.returncode == 2
    assert e.code == 2
    with pytest.raises(AttributeError):
        assert e.xyz == 3
    with pytest.raises(AttributeError):
        e._aliases = 1



from ..patch import patched_env
from os import environ


def test_patched_env():
    if 'HOME' in environ:
        home = environ['HOME']
        with patched_env(HOME=None):
            assert 'HOME' not in environ
        assert environ['HOME'] == home
    unusual_name = 'DATALADPATCHENVTESTVAR'
    if unusual_name not in environ:
        with patched_env(**{unusual_name: 'dummy'}):
            assert environ[unusual_name] == 'dummy'
        assert unusual_name not in environ



"""Enable reproducible documentation

The original documentation generation tooling can include memory
addresses in the text and thereby prevent reproducible builds
with unnecessary noise.

This patch uses regex to look for and delete such information
in a similar way to what sphinx is doing anyway in other places.
"""

import re

from datalad.interface.base import (
    alter_interface_docs_for_api,
    is_api_arg,
)

from . import apply_patch

obj_address_repr = re.compile(' at 0x[0-9a-f]+')


# This function interface is initially taken from
# datalad-core@982cca549ae29a1c86a0d6736bc3d6dfec370433
def update_docstring_with_parameters(func, params, prefix=None, suffix=None,
                                     add_args=None):
    """Generate a useful docstring from a parameter spec

    Amends any existing docstring of a callable with a textual
    description of its parameters. The Parameter spec needs to match
    the number and names of the callables arguments.
    """
    from datalad.utils import getargspec

    # get the signature
    args, varargs, varkw, defaults = getargspec(func, include_kwonlyargs=True)
    defaults = defaults or tuple()
    if add_args:
        add_argnames = sorted(add_args.keys())
        args.extend(add_argnames)
        defaults = defaults + tuple(add_args[k] for k in add_argnames)
    ndefaults = len(defaults)
    # start documentation with what the callable brings with it
    doc = prefix if prefix else u''
    if len(args) > 1:
        if len(doc):
            if not doc.endswith('\n'):
                doc += '\n'
            doc += '\n'
        doc += "Parameters\n----------\n"
        for i, arg in enumerate(args):
            if not is_api_arg(arg):
                continue
            # we need a parameter spec for each argument
            if not arg in params:
                raise ValueError("function has argument '%s' not described as a parameter" % arg)
            param = params[arg]
            # validate the default -- to make sure that the parameter description is
            # somewhat OK
            defaults_idx = ndefaults - len(args) + i
            if defaults_idx >= 0:
                if param.constraints is not None:
                    param.constraints(defaults[defaults_idx])
            orig_docs = param._doc
            param._doc = alter_interface_docs_for_api(param._doc)
            autodoc = param.get_autodoc(
                arg,
                default=defaults[defaults_idx] if defaults_idx >= 0 else None,
                has_default=defaults_idx >= 0)
            # PATCH replace any memory address in a repr()-like report
            # with nothing to make doc generation reproducible
            doc += obj_address_repr.sub('', autodoc)
            param._doc = orig_docs
            doc += '\n'
    doc += suffix if suffix else u""
    # assign the amended docs
    func.__doc__ = doc
    return func


apply_patch(
    'datalad.interface.base', None, 'update_docstring_with_parameters',
    update_docstring_with_parameters,
)



"""This file collects all patches for ORA/RIA-related code.

The patches have to goals:

1. Improve stability and consolidate code by using persistent shell support in
   class :class:`SSHRemoteIO`.

2. Improve ORA/RIA-related code so that it also works on Windows.
"""

from . import (
    add_method_url2transport_path,
    # this replaces SSHRemoteIO entirely
    replace_sshremoteio,
    # The following patches add Windows-support to ORA/RIA code
    ria_utils,
    replace_ora_remote,
    # `replace_create_sibling_ria` be imported after `replace_sshremoteio`
    # and `ria_utils`.
    replace_create_sibling_ria,
)



"""A persistent shell connection

This module provides a context manager that establishes a connection to a shell
and can be used to execute multiple commands in that shell. Shells are usually
remote shells, e.g. connected via an ``ssh``-client, but local shells like
``zsh``, ``bash`` or ``PowerShell`` can also be used.

The context manager returns an instance of :class:`ShellCommandExecutor` that
can be used to execute commands in the shell via the method
:meth:`ShellCommandExecutor.__call__`. The method will return an instance of
a subclass of :class:`ShellCommandResponseGenerator` that can be used to
retrieve the output of the command, the result code of the command, and the
stderr-output of the command.

Every response generator expects a certain output structure. It is responsible
for ensuring that the output structure is generated. To this end every
response generator provides a method
:meth:`ShellCommandResponseGenerator.get_command_list`. The method
:class:`ShellCommandExecutor.__call__` will pass the user-provided command to
:meth:`ShellCommandResponseGenerator.get_command_list` and receive a list of
final commands that should be executed in the connected shell and that will
generate the expected output structure. Instances of
:class:`ShellCommandResponseGenerator` have therefore four tasks:

    1. Create a final command list that is used to execute the user provided
       command. This could, for example, execute the command, print an
       end marker, and print the return code of the command.

    2. Parse the output of the command, yield it to the user.

    3. Read the return code and provide it to the user.

    4. Provide stderr-output to the user.

A very versatile example of a response generator is the class
:class:`VariableLengthResponseGenerator`. It can be used to execute a command
that will result in an output of unknown length, e.g. ``ls``, and will yield
the output of the command to the user. It does that by using a random
*end marker* to detect the end of the output and read the trailing return code.
This is suitable for almost all commands.

If :class:`VariableLengthResponseGenerator` is so versatile, why not just
implement its functionality in :class:`ShellCommandExecutor`? There are two
major reasons for that:

1. Although the :class:`VariableLengthResponseGenerator` is very versatile,
   it is not the most efficient implementation for commands that produce large
   amounts of output. In addition, there is also a minimal risk that the end
   marker is part of the output of the command, which would trip up the response
   generator. Putting response generation into a separate class allows to
   implement specific operations more efficiently and more safely.
   For example,
   :class:`DownloadResponseGenerator` implements the download of files. It
   takes a remote file name as user "command" and creates a final command list
   that emits the length of the file, a newline, the file content, a return
   code, and a newline. This allows :class:`DownloadResponseGenerator`
   to parse the output without relying on an end marker, thus increasing
   efficiency and safety

2. Factoring out the response generation creates an interface that can be used
   to support the syntax of different shells and the difference in command
   names and options in different operating systems. For example, the response
   generator class :class:`VariableLengthResponseGeneratorPowerShell` supports
   the invocation of commands with variable length output in a ``PowerShell``.

In short, parser generator classes encapsulate details of shell-syntax and
operation implementation. That allows support of different shell syntax, and
the efficient implementation of specific higher level operations, e.g.
``download``. It also allows users to extend the functionality of
:class:`ShellCommandExecutor` by providing their own response generator
classes.

The module :mod:`datalad_next.shell.response_generators` provides two generally
applicable abstract response generator classes:

    - :class:`VariableLengthResponseGenerator`

    - :class:`FixedLengthResponseGenerator`

The functionality of the former is described above. The latter can be used to
execute a command that will result in output of known
length, e.g. ``echo -n 012345``. It reads the specified number of bytes and a
trailing return code. This is more performant than the variable length response
generator (because it does not have to search for the end marker). In addition,
it does not rely on the uniqueness of the end marker. It is most useful for
operation like ``download``, where the length of the output can be known in
advance.

As mentioned above, the classes :class:`VariableLengthResponseGenerator` and
:class:`FixedLengthResponseGenerator` are abstract. The module
:mod:`datalad_next.shell.response_generators` provides the following concrete
implementations for them:

    - :class:`VariableLengthResponseGeneratorPosix`

    - :class:`VariableLengthResponseGeneratorPowerShell`

    - :class:`FixedLengthResponseGeneratorPosix`

    - :class:`FixedLengthResponseGeneratorPowerShell`

When :func:`datalad_next.shell.shell` is executed it will use a
:class:`VariableLengthResponseClass` to skip the login message of the shell.
This is done by executing a *zero command* (a command that will possibly
generate some output, and successfully return) in the shell. The zero command is
provided by the concrete implementation of class
:class:`VariableLengthResponseGenerator`. For example, the zero command for
POSIX shells is ``test 0 -eq 0``, for PowerShell it is ``Write-Host hello``.

Because there is no way for func:`shell` to determine the kind of shell it
connects to, the user can provide an alternative response generator class, in
the ``zero_command_rg_class``-parameter. Instance of that class
will then be used to execute the zero command. Currently, the following two
response generator classes are available:

    - :class:`VariableLengthResponseGeneratorPosix`: works with POSIX-compliant
      shells, e.g. ``sh`` or ``bash``. This is the default.
    - :class:`VariableLengthResponseGeneratorPowerShell`: works with PowerShell.

Whenever a command is executed via :meth:`ShellCommandExecutor.__call__`, the
class identified by ``zero_command_rg_class`` will be used by default to create
the final command list and to parse the result. Users can override this on a
per-call basis by providing a different response generator class in the
``response_generator``-parameter of :meth:`ShellCommandExecutor.__call__`.

Examples
--------

See the documentation of :func:`datalad_next.shell.shell` for examples of how to
use the shell-function and different response generator classes.

API overview
------------
.. currentmodule:: datalad_next.shell

.. autosummary::
   :toctree: generated
   :recursive:

   shell
   ShellCommandExecutor
   ShellCommandResponseGenerator
   VariableLengthResponseGenerator
   VariableLengthResponseGeneratorPosix
   VariableLengthResponseGeneratorPowerShell
   FixedLengthResponseGenerator
   FixedLengthResponseGeneratorPosix
   FixedLengthResponseGeneratorPowerShell
   DownloadResponseGenerator
   DownloadResponseGeneratorPosix
   operations.posix.upload
   operations.posix.download
   operations.posix.delete
"""


__all__ = [
    'shell',
    'posix',
]

from .shell import (
    shell,
    ShellCommandExecutor,
)

from .operations import posix
from .operations.posix import (
    DownloadResponseGenerator,
    DownloadResponseGeneratorPosix,
)
from .response_generators import (
    FixedLengthResponseGenerator,
    FixedLengthResponseGeneratorPosix,
    FixedLengthResponseGeneratorPowerShell,
    ShellCommandResponseGenerator,
    VariableLengthResponseGenerator,
    VariableLengthResponseGeneratorPosix,
    VariableLengthResponseGeneratorPowerShell,
)



from pathlib import Path
import pytest
from typing import Generator
from unittest.mock import (
    MagicMock,
    call,
    patch,
)

from datalad_next.tests import (
    assert_in,
    assert_in_results,
    eq_,
)

from datalad_next.patches.push_to_export_remote import (
    _get_export_log_entry,
    _is_export_remote,
    _is_valid_treeish,
    _transfer_data,
    get_export_records,
    mod_push,
)


module_name = "datalad_next.patches.push_to_export_remote"


class MockRepo:
    def __init__(self, return_special_remotes: bool = True):
        self.return_special_remotes = return_special_remotes

    def get_special_remotes(self):
        if self.return_special_remotes:
            return {
                0: {
                    "name": "no-target",
                    "exporttree": "no"
                },
                1: {
                    "name": "yes-target",
                    "exporttree": "yes"
                },
                2: {
                    "name": "some-target",
                    "exporttree": "no"
                }
            }
        else:
            return {}

    def call_git(self, *args, **kwargs):
        return

    def _call_annex_records_items_(self, *args, **kwargs):
        yield {
            "command": f"export {args[0][3]}",
            "file": "file.txt",
            "success": True,
            "input": [],
            "error-messages": []
        }
        yield {
            "command": f"export {args[0][3]}",
            "success": False,
            "input": [],
            "error-messages":
                ["external special remote error: WHATEVER WENT WRONG"],
            "file": "somefile"}


def _call_transfer(target: str,
                   config_result: bool,
                   return_special_remotes: bool = True) -> Generator:
    ds_mock = MagicMock()
    ds_mock.config.getbool.return_value = config_result
    ds_mock.pathobj = Path("/root")
    return _transfer_data(
        repo=MockRepo(return_special_remotes),
        ds=ds_mock,
        target=target,
        content=[],
        data="",
        force=None,
        jobs=None,
        res_kwargs={"path": str(Path("/root"))},
        got_path_arg=False)


def test_is_export_remote():
    # Ensure that None is handled properly
    assert not _is_export_remote(None)

    # Ensure that dicts without "exporttree" keyword are handled correctly
    assert not _is_export_remote({})

    # Ensure that "exporttree" is interpreted correctly
    assert not _is_export_remote({"exporttree": "no"})
    assert _is_export_remote({"exporttree": "yes"})


def test_patch_pass_through():
    # Ensure that the original _transfer_data is called if the target remote
    # has exporttree # not set to "yes"
    with patch("datalad_next.patches.push_to_export_remote.mod_push._push_data") as pd_mock:
        tuple(_call_transfer("no-target", False))
        eq_(pd_mock.call_count, 1)


def test_patch_execute_export():
    # Ensure that export is called if the target remote has exporttree set to
    # "yes"
    with patch(f"{module_name}.mod_push._push_data") as pd_mock, \
         patch(f"{module_name}._get_export_log_entry") as gele_mock:
        gele_mock.return_value = None
        results = tuple(_call_transfer("yes-target", False))
        eq_(pd_mock.call_count, 0)
        assert_in_results(results,
                          path=str(Path("/root/file.txt")),
                          target="yes-target",
                          action="copy",
                          status="ok")
        assert_in_results(results,
                          path=str(Path("/root/somefile")),
                          target="yes-target",
                          action="copy",
                          status="error")


def test_patch_skip_ignore_targets_export():
    with patch(f"{module_name}.lgr") as lgr_mock:
        tuple(_call_transfer("yes-target", True))
        assert_in(
            call.debug(
                "Target '%s' is set to annex-ignore, exclude from data-export.",
                'yes-target'
            ),
            lgr_mock.mock_calls
        )


def test_patch_check_envpatch():
    # Ensure that export is called if the target remote has exporttree not set
    # to "yes"
    with patch(f"{module_name}.mod_push._push_data") as pd_mock, \
         patch(f"{module_name}.needs_specialremote_credential_envpatch") as nsce_mock, \
         patch(f"{module_name}.get_specialremote_credential_envpatch") as gsce_mock, \
         patch(f"{module_name}._get_export_log_entry") as gele_mock, \
         patch(f"{module_name}._get_credentials") as gc_mock:

        nsce_mock.return_value = True
        gsce_mock.return_value = {"WEBDAVU": "hans", "WEBDAVP": "abc"}
        gele_mock.return_value = None
        gc_mock.return_value = {"secret": "abc", "user": "hans"}
        results = tuple(_call_transfer("yes-target", False))
        eq_(pd_mock.call_count, 0)
        assert_in_results(results,
                          path=str(Path("/root/file.txt")),
                          target="yes-target",
                          action="copy",
                          status="ok")
        assert_in_results(results,
                          path=str(Path("/root/somefile")),
                          target="yes-target",
                          action="copy",
                          status="error")


def test_no_special_remotes():
    # Ensure that the code works if no special remotes exist
    with patch(f"{module_name}.mod_push._push_data") as pd_mock:
        tuple(_call_transfer("no-target", False, False))
        eq_(pd_mock.call_count, 1)


def test_get_export_records_no_exports():
    class NoExportRepo:
        def call_git_items_(self, *args, **kwargs):
            raise mod_push.CommandError(
                stderr="fatal: Not a valid object name git-annex:export.log")

    results = tuple(get_export_records(NoExportRepo()))
    eq_(results, ())


def test_get_export_records():
    class SomeExportsRepo:
        def call_git_items_(self, *args, **kwargs):
            return [
                f"{i}.3s from{i}:to 0000{i}"
                for i in (3, 1, 4, 5, 2)
            ]

    result = tuple(get_export_records(SomeExportsRepo()))
    expected = tuple(
        {
            "timestamp": float(i + .3),
            "source-annex-uuid": f"from{i}",
            "destination-annex-uuid": f"to",
            "treeish": f"0000{i}"
        }
        for i in range(1, 6)
    )
    for remote_info in expected:
        assert_in(remote_info, result)


def test_get_export_log_entry():
    # Expect the youngest entry to be returned.
    class ManyExportsRepo:
        def call_git_items_(self, *args, **kwargs):
            return [
                f"{i}.3s from{i}:to 0000{i}"
                for i in (3, 4, 1, 5, 2)
            ]

        def get_export_records(self):
            yield from get_export_records(self)

    result = _get_export_log_entry(ManyExportsRepo(), "to")
    eq_(
        result,
        {
            "timestamp": 5.3,
            "source-annex-uuid": "from5",
            "destination-annex-uuid": "to",
            "treeish": f"00005"
        }
    )


def test_is_valid_treeish():
    pytest.skip(
        "this test is skipped until issue "
        "https://github.com/datalad/datalad-next/issues/39 is solved")

    class LogRepo:
        def call_git_items_(self, *args, **kwargs):
            return [
                f"commit{i} 0000{i}"
                for i in range(4)
            ]

    # Check successful validation
    export_entry = {"treeish": "00002"}
    assert _is_valid_treeish(LogRepo(), export_entry)

    # Check unsuccessful validation
    export_entry = {"treeish": "10000"}
    assert not _is_valid_treeish(LogRepo(), export_entry)



"""python-requests-compatible authentication handler using DataLad credentials
"""
# allow for |-type UnionType declarations
from __future__ import annotations

import logging
from typing import Dict
from urllib.parse import urlparse

import requests

from datalad_next.config import ConfigManager
from datalad_next.credman import CredentialManager

from .http_helpers import get_auth_realm

lgr = logging.getLogger('datalad.ext.next.utils.requests_auth')


__all__ = ['DataladAuth', 'HTTPBearerTokenAuth', 'parse_www_authenticate']


def parse_www_authenticate(hdr: str) -> dict:
    """Parse HTTP www-authenticate header

    This helper uses ``requests`` utilities to parse the ``www-authenticate``
    header as represented in a ``requests.Response`` instance. The header may
    contain any number of challenge specifications.

    The implementation follows RFC7235, where a challenge parameters set is
    specified as: either a comma-separated list of parameters, or a single
    sequence of characters capable of holding base64-encoded information,
    and parameters are name=value pairs, where the name token is matched
    case-insensitively, and each parameter name MUST only occur once
    per challenge.

    Returns
    -------
    dict
      Keys are casefolded challenge labels (e.g., 'basic', 'digest').
      Values are: ``None`` (no parameter), ``str`` (a token68), or
      ``dict`` (name/value mapping of challenge parameters)
    """
    plh = requests.utils.parse_list_header
    pdh = requests.utils.parse_dict_header
    challenges = {}
    challenge = None
    # challenges as well as their properties are in a single
    # comma-separated list
    for item in plh(hdr):
        # parse the item into a key/value set
        # the value will be `None` if this item was no mapping
        k, v = pdh(item).popitem()
        # split the key to check for a challenge spec start
        key_split = k.split(' ', maxsplit=1)
        if len(key_split) > 1 or v is None:
            item_suffix = item[len(key_split[0]) + 1:]
            challenge = [item[len(key_split[0]) + 1:]] if item_suffix else None
            challenges[key_split[0].casefold()] = challenge
        else:
            # implementation logic assumes that the above conditional
            # was triggered before we ever get here
            assert challenge
            challenge.append(item)

    return {
        challenge: _convert_www_authenticate_items(items)
        for challenge, items in challenges.items()
    }


def _convert_www_authenticate_items(items: list) -> None | str | dict:
    pdh = requests.utils.parse_dict_header
    # according to RFC7235, items can be:
    # either a comma-separated list of parameters
    # or a single sequence of characters capable of holding base64-encoded
    # information.
    # parameters are name=value pairs, where the name token is matched
    # case-insensitively, and each parameter name MUST only occur once
    # per challenge.
    if items is None:
        return None
    elif len(items) == 1 and pdh(items[0].rstrip('=')).popitem()[1] is None:
        # this items matches the token68 appearance (no name value
        # pair after potential base64 padding its removed
        return items[0]
    else:
        return {
            k.casefold(): v for i in items for k, v in pdh(i).items()
        }


class DataladAuth(requests.auth.AuthBase):
    """Requests-style authentication handler using DataLad credentials

    Similar to request_toolbelt's `AuthHandler`, this is a meta
    implementation that can be used with different actual authentication
    schemes. In contrast to `AuthHandler`, a credential can not only be
    specified directly, but credentials can be looked up based on the
    target URL and the server-supported authentication schemes.

    In addition to programmatic specification and automated lookup, manual
    credential entry using interactive prompts is also supported.

    At present, this implementation is not thread-safe.
    """
    _supported_auth_schemes = {
        'basic': 'user_password',
        'digest': 'user_password',
        'bearer': 'token',
    }

    def __init__(self, cfg: ConfigManager, credential: str | None = None):
        """
        Parameters
        ----------
        cfg: ConfigManager
          Is passed to CredentialManager() as `cfg`-parameter.
        credential: str, optional
          Name of a particular credential to be used for any operations.
        """
        self._credman = CredentialManager(cfg)
        self._credential = credential
        self._entered_credential = None

    def save_entered_credential(self, suggested_name: str | None = None,
                                context: str | None = None) -> Dict | None:
        """Utility method to save a pending credential in the store

        Pending credentials have been entered manually, and were subsequently
        used successfully for authentication.

        Saving a credential will prompt for entering a name to identify the
        credentials.
        """
        if self._entered_credential is None:
            # nothing to do
            return None
        return self._credman.set(
            name=None,
            _lastused=True,
            _suggested_name=suggested_name,
            _context=context,
            **self._entered_credential
        )

    def __call__(self, r):
        # TODO support being called from multiple threads
        #self.init_per_thread_state()

        # register hooks to be executed from a response to this
        # request is available
        # Redirect: reset credentials to avoid leakage to other server
        r.register_hook("response", self.handle_redirect)
        # 401 Unauthorized: look for a credential and try again
        r.register_hook("response", self.handle_401)
        return r

    def _get_credential(self, url, auth_schemes
                        ) -> tuple[str | None, str | None, Dict | None]:
        """Get a credential for access to `url` given server-supported schemes

        If a particular credential to use was given to the `DataladAuth`
        constructor it reported here.

        In all other situations a credential will be looked up, based on
        the access URL and the authentication schemes supported by the
        host. The authentication schemes will be tested in the order in
        which they are reported by the remote host.

        If no matching credential can be identified, a prompt to enter
        a credential is presented. The credential type will match, and be
        used with the first authentication scheme that is both reported
        by the host, and by this implementation.

        The methods returns a 3-tuple. The first element is an identifier
        for the authentication scheme ('basic', digest', etc.) to use
        with the credential. The second item is the name for the reported
        credential, and the third element is a dictionary with the
        credential properties and its secret. Any of the three items can be
        `None` if the respective information is not available.
        """
        if self._credential:
            cred = self._credman.get(name=self._credential)
            # this credential is scheme independent
            return None, self._credential, cred

        # no credential identified, find one
        for ascheme in auth_schemes:
            if ascheme not in DataladAuth._supported_auth_schemes:
                # nothing we can handle
                continue
            ctype = DataladAuth._supported_auth_schemes[ascheme]
            # get a realm ID for this authentication scheme
            realm = get_auth_realm(url, auth_schemes, scheme=ascheme)
            # ask for matching credentials
            creds = [
                (name, cred) for name, cred in self._credman.query(
                    _sortby='last-used',
                    type=ctype,
                    realm=realm,
                )
                # we can only work with complete credentials, although
                # query() might return others. We exclude them here
                # to be able to fall back on manual entry further down
                if cred.get('secret')
            ]
            if creds:
                # we have matches, go with the last used one
                name, cred = creds[0]
                return ascheme, name, cred

        # no success finding an existing credential, now ask, if possible
        # pick a scheme that is supported by the server and by us
        ascheme = [s for s in auth_schemes
                   if s in DataladAuth._supported_auth_schemes]
        if not ascheme:
            # f-string OK, only executed on failure
            lgr.debug(
                'Only unsupported HTTP auth schemes offered '
                f'{list(auth_schemes.keys())!r}')
        # go with the first supported scheme
        ascheme = ascheme[0]
        ctype = DataladAuth._supported_auth_schemes[ascheme]

        try:
            realm = get_auth_realm(url, auth_schemes)
            cred = self._credman.get(
                name=None,
                _prompt=f'Credential needed for accessing {url} (authentication realm {realm!r})',
                _type_hint=ctype,
                type=ctype,
                # include the realm in the credential to avoid asking for it
                # interactively (it is a server-specified property
                # users would generally not know, if they do, they can use the
                # `credentials` command upfront.
                realm=realm
            )
            self._entered_credential = cred
            return ascheme, None, cred
        except Exception as e:
            lgr.debug('Credential retrieval failed: %s', e)
            return ascheme, None, None

    def handle_401(self, r, **kwargs):
        """Callback that received any response to a request

        Any non-4xx response or a response lacking a 'www-authenticate'
        header is ignored.

        Server-provided 'www-authenticated' challenges are inspected, and
        corresponding credentials are looked-up (if needed) and subsequently
        tried in a re-request to the original URL after performing any
        necessary actions to meet a given challenge. Such a re-request
        is then using the same connection as the original request.

        Particular challenges are implemented in dedicated classes, e.g.
        :class:`requests.auth.HTTPBasicAuth`.

        Credential look-up or entry is performed by
        :meth:`datalad_next.requests_auth.DataladAuth._get_credential`.
        """
        if not 400 <= r.status_code < 500:
            # fast return if this is no error, see
            # https://github.com/psf/requests/issues/3772 for background
            return r
        if 'www-authenticate' not in r.headers:
            # no info on how to authenticate to react to, leave it as-is.
            # this also catches any non-401-like error code (e.g. 429).
            # doing this more loose check (rather then going for 401
            # specifically) enables to support services that send
            # www-authenticate with e.g. 403s
            return r
        # which auth schemes does the server support?
        auth_schemes = parse_www_authenticate(r.headers['www-authenticate'])
        ascheme, credname, cred = self._get_credential(r.url, auth_schemes)

        if cred is None or 'secret' not in cred:
            # we got nothing, leave things as they are
            return r

        # TODO add safety check. if a credential somehow contains
        # information on its scope (i.e. only for github.com)
        # prevent its use for other hosts -- maybe unless given explicitly.

        if ascheme is None:
            # if there is no authentication scheme identified, look at the
            # credential, if it knows
            ascheme = cred.get('http_auth_scheme')
        if ascheme is None:
            # go with the first supported scheme that matches the credential
            # type; fall back to 'bearer' for token credentials when the
            # server advertises only basic (e.g. dCache with empty realm)
            matching = [
                c for c in auth_schemes
                if c in DataladAuth._supported_auth_schemes
                and cred.get('type') == DataladAuth._supported_auth_schemes[c]
            ]
            if matching:
                ascheme = matching[0]
            elif cred.get('type') == 'token':
                ascheme = 'bearer'

        if ascheme == 'basic':
            return self._authenticated_rerequest(
                r,
                requests.auth.HTTPBasicAuth(cred['user'], cred['secret']),
                **kwargs)
        elif ascheme == 'digest':
            return self._authenticated_rerequest(
                r,
                requests.auth.HTTPDigestAuth(cred['user'], cred['secret']),
                **kwargs)
        elif ascheme == 'bearer':
            return self._authenticated_rerequest(
                r, HTTPBearerTokenAuth(cred['secret']), **kwargs)
        else:
            raise NotImplementedError(
                'Only unsupported HTTP auth schemes offered '
                f'{list(auth_schemes.keys())!r} need {ascheme!r}')

    def handle_redirect(self, r, **kwargs):
        """Callback that received any response to a request

        Any non-redirect response is ignore.

        This callback drops an explicitly set credential whenever
        the redirect causes a non-encrypted connection to be used
        after the original request was encrypted, or when the `netloc`
        of the redirect differs from the original target.
        """
        if r.is_redirect and self._credential:
            og_p = urlparse(r.url)
            rd_p = urlparse(r.headers.get('location'), '')
            if og_p.netloc != rd_p.netloc or (
                    rd_p.scheme == 'http' and og_p.scheme == 'https'):
                lgr.debug(
                    'URL redirect, discarded given credential %r '
                    'to avoid leakage',
                    self._credential)
                self._credential = None

    def _authenticated_rerequest(
            self,
            response: requests.models.Response,
            auth: requests.auth.AuthBase,
            **kwargs
    ) -> requests.models.Response:
        """Helper to rerun a request, but with authentication added"""
        prep = _get_renewed_request(response)
        auth(prep)
        _r = response.connection.send(prep, **kwargs)
        _r.history.append(response)
        _r.request = prep
        return _r


def _get_renewed_request(r: requests.models.Response
                         ) -> requests.models.PreparedRequest:
    """Helper. Logic taken from requests.auth.HTTPDigestAuth"""
    # Consume content and release the original connection
    # to allow our new request to reuse the same one.
    r.content
    r.close()
    prep = r.request.copy()
    requests.cookies.extract_cookies_to_jar(
        prep._cookies, r.request, r.raw)
    prep.prepare_cookies(prep._cookies)
    return prep


class HTTPBearerTokenAuth(requests.auth.AuthBase):
    """Attaches HTTP Bearer Token Authentication to the given Request object.
    """
    def __init__(self, token):
        super().__init__()
        self.token = token

    def __call__(self, r):
        r.headers["Authorization"] = f'Bearer {self.token}'
        return r



import pytest

from datalad_next.runners import call_git

from .. import get_worktree_head


def test_get_worktree_head(tmp_path, existing_dataset):
    ds = existing_dataset

    with pytest.raises(ValueError) as e:
        get_worktree_head(tmp_path / 'IDONOTEXISTONTHEFILESYSTEM')
    assert str(e.value) == 'path not found'

    norepo = tmp_path / 'norepo'
    norepo.mkdir()
    with pytest.raises(ValueError) as e:
        get_worktree_head(norepo)
    assert str(e.value) == f'no Git repository at {norepo!r}'

    reponohead = tmp_path / 'reponohead'
    reponohead.mkdir()
    call_git(['init'], cwd=reponohead)
    assert (None, None) == get_worktree_head(reponohead)

    # and actual repo with a commit
    head, chead = get_worktree_head(ds.pathobj)
    # we always get a HEAD
    # we always get fullname symbolic info
    assert head.startswith('refs/heads/')
    if chead is not None:
        # there is a corresponding head, and we get it as the
        # git-annex 'basis' ref
        assert head.startswith('refs/heads/adjusted/')
        assert chead.startswith('refs/basis/')



from __future__ import annotations

from pathlib import Path
from typing import (
    Iterable,
    List,
)
from datasalad.runners import CommandError as SaladCommandError
from datasalad.iterable_subprocess import iterable_subprocess

from datalad_next.exceptions import CommandError
from datalad_next.consts import COPY_BUFSIZE

__all__ = ['iter_subproc']


def iter_subproc(
    args: List[str],
    *,
    input: Iterable[bytes] | None = None,
    chunk_size: int = COPY_BUFSIZE,
    cwd: Path | None = None,
    bufsize: int = -1,
):
    """Context manager to communicate with a subprocess using iterables

    .. deprecated:: 1.6

       Use ``datasalad.runners.iter_proc`` instead. Renamed ``input`` argument
       to ``inputs``, and raises datalad's ``CommandError``.

    This offers a higher level interface to subprocesses than Python's
    built-in ``subprocess`` module. It allows a subprocess to be naturally
    placed in a chain of iterables as part of a data processing pipeline.
    It is also helpful when data won't fit in memory and has to be streamed.

    This is a convenience wrapper around ``datalad_next.iterable_subprocess``,
    which itself is a slightly modified (for use on Windows) fork of
    https://github.com/uktrade/iterable-subprocess, written by
    Michal Charemza.

    This function provides a context manager.
    On entering the context, the subprocess is started, the thread to read
    from standard error is started, the thread to populate subprocess
    input is started.
    When running, the standard input thread iterates over the input,
    passing chunks to the process, while the standard error thread
    fetches the error output, and while the main thread iterates over
    the process's output from client code in the context.

    On context exit, the main thread closes the process's standard output,
    waits for the standard input thread to exit, waits for the standard error
    thread to exit, and wait for the process to exit. If the process exited
    with a non-zero return code, a ``CommandError`` is raised,
    containing the process's return code.

    If the context is exited due to an exception that was raised in the
    context, the main thread terminates the process via ``Popen.terminate()``,
    closes the process's standard output, waits for the standard input
    thread to exit, waits for the standard error thread to exit, waits
    for the process to exit, and re-raises the exception.

    Note, if an exception is raised in the context, this exception will bubble
    up to the main thread. That means no ``CommandError`` will
    be raised if the subprocess exited with a non-zero return code.
    To access the return code in case of an exception inside the context,
    use the ``code``-attribute of the ``as``-variable.
    This object will always contain the return code of the subprocess.
    For example, the following code will raise a ``StopIteration``-exception
    in the context (by repeatedly using :func:`next`). The subprocess
    will exit with ``2`` due to the illegal option ``-@``, and no
    ``CommandError`` is raised. The return code is read from
    the variable ``ls_stdout``

    .. code-block:: python

      >>> from datalad_next.runners import iter_subproc
      >>> try:
      ...     with iter_subproc(['ls', '-@']) as ls_stdout:
      ...         while True:
      ...             next(ls_stdout)
      ... except Exception as e:
      ...     print(repr(e), ls_stdout.returncode)
      StopIteration() 2


    Parameters
    ----------
    args: list
      Sequence of program arguments to be passed to ``subprocess.Popen``.
    input: iterable, optional
      If given, chunks of ``bytes`` to be written, iteratively, to the
      subprocess's ``stdin``.
    chunk_size: int, optional
      Size of chunks to read from the subprocess's stdout/stderr in bytes.
    cwd: Path
      Working directory for the subprocess, passed to ``subprocess.Popen``.
    bufsize: int, optional
      Buffer size to use for the subprocess's ``stdin``, ``stdout``, and
      ``stderr``. See ``subprocess.Popen`` for details.

    Returns
    -------
    contextmanager
    """
    try:
        return iterable_subprocess(
            args,
            tuple() if input is None else input,
            chunk_size=chunk_size,
            cwd=cwd,
            bufsize=bufsize,
        )
    except SaladCommandError as e:
        raise CommandError(
            cmd=e.cmd,
            msg=e.msg,
            code=e.returncode,
            stdout=e.stdout,
            stderr=e.stderr,
            cwd=e.cwd,
        ) from e



import pytest

from ..annexkey import AnnexKey


def test_annexkey():
    for key in (
        'MD5E-s792207360--985e680a221e47db05063a12b91d7d89.tar',
        'SHA256E-s31390--f50d7ac4c6b9031379986bc362fcefb65f1e52621ce1708d537e740fefc59cc0.mp3',
        'URL-s1899248--http&c%%ai.stanford.edu%,126nilsson%MLBOOK.pdf/URL-s1899248--http&c%%ai.stanford.edu%,126nilsson%MLBOOK.pdf',
    ):
        # round-tripping for any key must give same outcome
        assert key == str(AnnexKey.from_str(key))

    # check that it can be used as a dict-key, i.e. is hashable
    key = AnnexKey.from_str('MD5-s9--985e680a221e47db05063a12b91d7d89')
    d = {key: 'some'}


def test_annexkey_errors():
    for wrong in (
        'MD5E-985e680a221e47db05063a12b91d7d89.tar',
        'MD5E-SUPRISE--985e680a221e47db05063a12b91d7d89.tar',
    ):
        with pytest.raises(ValueError):
            AnnexKey.from_str(wrong)



from __future__ import annotations

import pytest

from ..replace_ora_remote import (
    canonify_url,
    de_canonify_url,
)


@pytest.mark.parametrize("scheme", ['ria+file', 'file'])
def test_canonify(scheme, monkeypatch):
    url_uncanonified = scheme + '://C:/a/b/c'
    url_canonified = scheme + ':///C:/a/b/c'

    monkeypatch.setattr(
        'datalad_next.patches.replace_ora_remote.on_windows',
        True,
    )
    assert canonify_url(url_canonified) == url_canonified
    assert canonify_url(url_uncanonified) == url_canonified

    monkeypatch.setattr(
        'datalad_next.patches.replace_ora_remote.on_windows',
        False,
    )
    assert canonify_url(url_canonified) == url_canonified
    assert canonify_url(url_uncanonified) == url_uncanonified


@pytest.mark.parametrize("scheme", ['ria+file', 'file'])
def test_de_canonify(scheme, monkeypatch):
    url_uncanonified = scheme + '://C:/a/b/c'
    url_canonified = scheme + ':///C:/a/b/c'

    monkeypatch.setattr(
        'datalad_next.patches.replace_ora_remote.on_windows',
        True,
    )
    assert de_canonify_url(url_canonified) == url_uncanonified
    assert de_canonify_url(url_uncanonified) == url_uncanonified

    monkeypatch.setattr(
        'datalad_next.patches.replace_ora_remote.on_windows',
        False,
    )
    assert de_canonify_url(url_canonified) == url_canonified
    assert de_canonify_url(url_uncanonified) == url_uncanonified



from __future__ import annotations

from os import environ
from typing import (
    Dict,
    Mapping,
    Tuple,
)


def get_gitconfig_items_from_env() -> Mapping[str, str | Tuple[str, ...]]:
    """Parse git-config ENV (``GIT_CONFIG_COUNT|KEY|VALUE``) and return as dict

    This implementation does not use ``git-config`` directly, but aims to
    mimic its behavior with respect to parsing the environment as much
    as possible.

    Raises
    ------
    ValueError
      Whenever ``git-config`` would also error out, and includes an
      message in the respective exception that resembles ``git-config``'s
      for that specific case.

    Returns
    -------
    dict
      Configuration key-value mappings. When a key is declared multiple
      times, the respective values are aggregated in reported as a tuple
      for that specific key.
    """
    items: Dict[str, str | Tuple[str, ...]] = {}
    for k, v in ((_get_gitconfig_var_from_env(i, 'key'),
                  _get_gitconfig_var_from_env(i, 'value'))
                 for i in range(_get_gitconfig_itemcount())):
        val = items.get(k)
        if val is None:
            items[k] = v
        elif isinstance(val, tuple):
            items[k] = val + (v,)
        else:
            items[k] = (val, v)
    return items


def _get_gitconfig_itemcount() -> int:
    try:
        return int(environ.get('GIT_CONFIG_COUNT', '0'))
    except (TypeError, ValueError) as e:
        raise ValueError("bogus count in GIT_CONFIG_COUNT") from e


def _get_gitconfig_var_from_env(nid: int, kind: str) -> str:
    envname = f'GIT_CONFIG_{kind.upper()}_{nid}'
    var = environ.get(envname)
    if var is None:
        raise ValueError(f"missing config {kind} {envname}")
    if kind != 'key':
        return var
    if not var:
        raise ValueError(f"empty config key {envname}")
    if '.' not in var:
        raise ValueError(f"key {envname} does not contain a section: {var}")
    return var


def set_gitconfig_items_in_env(items: Mapping[str, str | Tuple[str, ...]]):
    """Set git-config ENV (``GIT_CONFIG_COUNT|KEY|VALUE``) from a mapping

    Any existing declaration of configuration items in the environment is
    replaced. Any ENV variable of a *valid* existing declaration is removed,
    before the set configuration items are posted in the ENV.

    Multi-value configuration keys are supported (values provided as a tuple).

    Any item with a value of ``None`` will be posted into the ENV with an
    empty string as value, i.e. the corresponding ``GIT_CONFIG_VALUE_{count}``
    variable will be an empty string. ``None`` item values indicate that the
    configuration key was unset on the command line, via the global option
    ``-c``.

    No verification (e.g., of syntax compliance) is performed.
    """
    _clean_env_from_gitconfig_items()

    count = 0
    for key, value in items.items():
        # homogeneous processing of multiple value items, and single values
        values = value if isinstance(value, tuple) else (value,)
        for v in values:
            environ[f'GIT_CONFIG_KEY_{count}'] = key
            # we support None even though not an allowed input type, because
            # of https://github.com/datalad/datalad/issues/7589
            # this can be removed, when that issue is resolved.
            environ[f'GIT_CONFIG_VALUE_{count}'] = '' if v is None else str(v)
            count += 1
    if count:
        environ['GIT_CONFIG_COUNT'] = str(count)


def _clean_env_from_gitconfig_items():
    # we only care about intact specifications here, if there was cruft
    # to start with, we have no responsibilities
    try:
        count = _get_gitconfig_itemcount()
    except ValueError:
        return

    for i in range(count):
        environ.pop(f'GIT_CONFIG_KEY_{i}', None)
        environ.pop(f'GIT_CONFIG_VALUE_{i}', None)

    environ.pop('GIT_CONFIG_COUNT', None)



import datalad_next.patches.enabled

# run all -core tests, because with _push() we patched a central piece
from datalad.core.distributed.tests.test_push import *

# import those directly to pass mypy tests, although they are already imported
# by the above import
from datalad.tests.utils_pytest import (
    DEFAULT_REMOTE,
    SkipTest,
    assert_in_results,
    assert_not_in_results,
    assert_result_count,
    eq_,
    known_failure_githubci_win,
    slow,
    with_tempfile,
)

from datalad_next.datasets import Dataset


# we override this specific test, because the original behavior is no longer
# value, because our implementation behaves "better"
def test_gh1811(tmp_path, no_result_rendering):
    srcpath = tmp_path / 'src'
    clonepath = tmp_path / 'clone'
    # `annex=false` is the only change from the -core implementation
    # of the test. For normal datasets with an annex, the problem underlying
    # gh1811 is no longer valid, because of more comprehensive analysis of
    # what needs pushing in this case
    orig = Dataset(srcpath).create(annex=False)
    (orig.pathobj / 'some').write_text('some')
    orig.save()
    clone = Clone.__call__(source=orig.path, path=clonepath)
    (clone.pathobj / 'somemore').write_text('somemore')
    clone.save()
    clone.repo.call_git(['checkout', 'HEAD~1'])
    res = clone.push(to=DEFAULT_REMOTE, on_failure='ignore')
    assert_result_count(res, 1)
    assert_result_count(
        res, 1,
        path=clone.path, type='dataset', action='publish',
        status='impossible',
        message='There is no active branch, cannot determine remote '
                'branch',
    )


# taken from datalad-core@250386f1fd83af7a3df72347c9b26a4afd66baa7
@slow  # can run over 30 sec when running in parallel with n=2. Cannot force serial yet, see https://github.com/pytest-dev/pytest-xdist/issues/385
# Removed @known_failure_githubci_win-decorator
@with_tempfile(mkdir=True)
@with_tempfile(mkdir=True)
@with_tempfile(mkdir=True)
def test_nested_pushclone_cycle_allplatforms(origpath=None, storepath=None, clonepath=None):
    if 'DATALAD_SEED' in os.environ:
        # we are using create-sibling-ria via the cmdline in here
        # this will create random UUIDs for datasets
        # however, given a fixed seed each call to this command will start
        # with the same RNG seed, hence yield the same UUID on the same
        # machine -- leading to a collision
        raise SkipTest(
            'Test incompatible with fixed random number generator seed'
        )
    # the aim here is this high-level test a std create-push-clone cycle for a
    # dataset with a subdataset, with the goal to ensure that correct branches
    # and commits are tracked, regardless of platform behavior and condition
    # of individual clones. Nothing fancy, just that the defaults behave in
    # sensible ways
    from datalad.cmd import WitlessRunner as Runner
    run = Runner().run

    os.environ['DATALAD_EXTENSIONS_LOAD'] = 'next'

    # create original nested dataset
    with chpwd(origpath):
        run(['datalad', 'create', 'super'], env=os.environ)
        run(
            [
                'datalad', 'create', '-d', 'super',
                str(Path('super', 'sub'))
            ],
            env=os.environ
        )

    # verify essential linkage properties
    orig_super = Dataset(Path(origpath, 'super'))
    orig_sub = Dataset(orig_super.pathobj / 'sub')

    (orig_super.pathobj / 'file1.txt').write_text('some1')
    (orig_sub.pathobj / 'file2.txt').write_text('some1')
    with chpwd(orig_super.path):
        run(['datalad', 'save', '--recursive'], env=os.environ)

    # TODO not yet reported clean with adjusted branches
    #assert_repo_status(orig_super.path)

    # the "true" branch that sub is on, and the gitsha of the HEAD commit of it
    orig_sub_corr_branch = \
        orig_sub.repo.get_corresponding_branch() or orig_sub.repo.get_active_branch()
    orig_sub_corr_commit = orig_sub.repo.get_hexsha(orig_sub_corr_branch)

    # make sure the super trackes this commit
    assert_in_results(
        orig_super.subdatasets(),
        path=orig_sub.path,
        gitshasum=orig_sub_corr_commit,
        # TODO it should also track the branch name
        # Attempted: https://github.com/datalad/datalad/pull/3817
        # But reverted: https://github.com/datalad/datalad/pull/4375
    )

    # publish to a store, to get into a platform-agnostic state
    # (i.e. no impact of an annex-init of any kind)
    store_url = 'ria+' + get_local_file_url(storepath)
    with chpwd(orig_super.path):
        run(
            [
                'datalad', 'create-sibling-ria', '--recursive',
                '-s', 'store', store_url, '--new-store-ok'
            ],
            env=os.environ
        )
        run(
            ['datalad', 'push', '--recursive', '--to', 'store'],
            env=os.environ
        )

    # we are using the 'store' sibling's URL, which should be a plain path
    store_super = AnnexRepo(orig_super.siblings(name='store')[0]['url'], init=False)
    store_sub = AnnexRepo(orig_sub.siblings(name='store')[0]['url'], init=False)

    # both datasets in the store only carry the real branches, and nothing
    # adjusted
    for r in (store_super, store_sub):
        eq_(set(r.get_branches()), set([orig_sub_corr_branch, 'git-annex']))

    # and reobtain from a store
    cloneurl = 'ria+' + get_local_file_url(str(storepath), compatibility='git')
    with chpwd(clonepath):
        run(
            ['datalad', 'clone', cloneurl + '#' + orig_super.id, 'super'],
            env=os.environ
        )
        run(
            ['datalad', '-C', 'super', 'get', '--recursive', '.'],
            env=os.environ
        )

    # verify that nothing has changed as a result of a push/clone cycle
    clone_super = Dataset(Path(clonepath, 'super'))
    clone_sub = Dataset(clone_super.pathobj / 'sub')
    assert_in_results(
        clone_super.subdatasets(),
        path=clone_sub.path,
        gitshasum=orig_sub_corr_commit,
    )

    for ds1, ds2, f in ((orig_super, clone_super, 'file1.txt'),
                        (orig_sub, clone_sub, 'file2.txt')):
        eq_((ds1.pathobj / f).read_text(), (ds2.pathobj / f).read_text())

    # get status info that does not recursive into subdatasets, i.e. not
    # looking for uncommitted changes
    # we should see no modification reported
    assert_not_in_results(
        clone_super.status(eval_subdataset_state='commit'),
        state='modified')
    # and now the same for a more expensive full status
    assert_not_in_results(
        clone_super.status(recursive=True),
        state='modified')



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See LICENSE file distributed along with the datalad_osf package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""'tree'-like command for visualizing dataset hierarchies"""

from __future__ import annotations

__docformat__ = "numpy"

import logging
from functools import (
    wraps,
    lru_cache
)
from os import readlink
from pathlib import Path

from datalad_next.commands import (
    EnsureCommandParameterization,
    ValidatedInterface,
    Parameter,
    build_doc,
    eval_results,
    datasetmethod,
    get_status_dict,
)
from datalad_next.exceptions import (
    CapturedException,
    NoDatasetFound
)

from datalad_next.iter_collections import iter_submodules
from datalad_next.constraints import (
    EnsureBool,
    EnsureDataset,
    EnsureInt,
    EnsurePath,
    EnsureRange,
)
from datalad_next.utils import get_dataset_root
from datalad_next.uis import ui_switcher as ui
from datalad_next.datasets import (
    Dataset,
)

lgr = logging.getLogger('datalad.local.tree')


@build_doc
class TreeCommand(ValidatedInterface):
    """Visualize directory and dataset hierarchies

    This command mimics the UNIX/MS-DOS 'tree' utility to generate and
    display a directory tree, with DataLad-specific enhancements.

    It can serve the following purposes:

    1. Glorified 'tree' command
    2. Dataset discovery
    3. Programmatic directory traversal

    *Glorified 'tree' command*

    The rendered command output uses 'tree'-style visualization::

        /tmp/mydir
        ├── [DS~0] ds_A/
        │   └── [DS~1] subds_A/
        └── [DS~0] ds_B/
            ├── dir_B/
            │   ├── file.txt
            │   ├── subdir_B/
            │   └── [DS~1] subds_B0/
            └── [DS~1] (not installed) subds_B1/

        5 datasets, 2 directories, 1 file

    Dataset paths are prefixed by a marker indicating subdataset hierarchy
    level, like ``[DS~1]``.
    This is the absolute subdataset level, meaning it may also take into
    account superdatasets located above the tree root and thus not included
    in the output.
    If a subdataset is registered but not installed (such as after a
    non-recursive ``datalad clone``), it will be prefixed by ``(not
    installed)``. Only DataLad datasets are considered, not pure
    git/git-annex repositories.

    The 'report line' at the bottom of the output shows the count of
    displayed datasets, in addition to the count of directories and
    files. In this context, datasets and directories are mutually
    exclusive categories.

    By default, only directories (no files) are included in the tree,
    and hidden directories are skipped. Both behaviours can be changed
    using command options.

    Symbolic links are always followed.
    This means that a symlink pointing to a directory is traversed and
    counted as a directory (unless it potentially creates a loop in
    the tree).

    *Dataset discovery*

    Using the [CMD: ``--recursive`` CMD][PY: ``recursive`` PY] or [CMD:
    ``--recursion-limit`` CMD][PY: ``recursion_limit`` PY]
    option, this command generates the layout of dataset hierarchies based on
    subdataset nesting level, regardless of their location in the
    filesystem.

    In this case, tree depth is determined by subdataset depth. This mode
    is thus suited for discovering available datasets when their
    location is not known in advance.

    By default, only datasets are listed, without their contents. If
    [CMD: ``--depth`` CMD][PY: ``depth`` PY] is specified additionally,
    the contents of each dataset will be included up to [CMD:
    ``--depth`` CMD][PY: ``depth`` PY] directory levels (excluding
    subdirectories that are themselves datasets).

    Tree filtering options such as [CMD: ``--include-hidden`` CMD][PY:
    ``include_hidden`` PY] only affect which directories are
    reported as dataset contents, not which directories are traversed to find
    datasets.

    **Performance note**: since no assumption is made on the location of
    datasets, running this command with the [CMD: ``--recursive`` CMD][PY:
    ``recursive`` PY] or [CMD: ``--recursion-limit`` CMD][PY:
    ``recursion_limit`` PY] option does a full scan of the whole directory
    tree. As such, it can be significantly slower than a call with an
    equivalent output that uses [CMD: ``--depth`` CMD][PY: ``depth`` PY] to
    limit the tree instead.

    *Programmatic directory traversal*

    The command yields a result record for each tree node (dataset,
    directory or file). The following properties are reported, where available:

    "path"
        Absolute path of the tree node

    "type"
        Type of tree node: "dataset", "directory" or "file"

    "depth"
        Directory depth of node relative to the tree root

    "exhausted_levels"
        Depth levels for which no nodes are left to be generated (the
        respective subtrees have been 'exhausted')

    "count"
        Dict with cumulative counts of datasets, directories and files in the
        tree up until the current node. File count is only included if the
        command is run with the [CMD: ``--include-files`` CMD][PY:
        ``include_files`` PY]
        option.

    "dataset_depth"
        Subdataset depth level relative to the tree root. Only included for
        node type "dataset".

    "dataset_abs_depth"
        Absolute subdataset depth level. Only included for node type "dataset".

    "dataset_is_installed"
        Whether the registered subdataset is installed. Only included for node
        type "dataset".

    "symlink_target"
        If the tree node is a symlink, the path to the link target

    "is_broken_symlink"
        If the tree node is a symlink, whether it is a broken symlink

    """

    result_renderer = 'tailored'

    _params_ = dict(
        path=Parameter(
            args=("path",),
            nargs='?',
            doc="""path to directory from which to generate the tree.
            Defaults to the current directory."""),
        depth=Parameter(
            args=("-L", "--depth",),
            doc="""limit the tree to maximum level of subdirectories.
            If not specified, will generate the full tree with no depth
            constraint.
            If paired with [CMD: ``--recursive`` CMD][PY: ``recursive`` PY] or
            [CMD: ``--recursion-limit`` CMD][PY: ``recursion_limit`` PY],
            refers to the maximum directory level to output below
            each dataset."""),
        recursive=Parameter(
            args=("-r", "--recursive",),
            doc="""produce a dataset tree of the full hierarchy of nested
            subdatasets. *Note*: may have slow performance on large
            directory trees.""",
            action='store_true'),
        recursion_limit=Parameter(
            args=("-R", "--recursion-limit",),
            metavar="LEVELS",
            doc="""limit the dataset tree to maximum level of nested
            subdatasets. 0 means include only top-level datasets, 1 means
            top-level datasets and their immediate subdatasets, etc. *Note*:
            may have slow performance on large directory trees."""),
        include_files=Parameter(
            args=("--include-files",),
            doc="""include files in the tree""",
            action='store_true'),
        include_hidden=Parameter(
            args=("--include-hidden",),
            doc="""include hidden files/directories in the tree. This
            option does not affect which directories will be searched for
            datasets when specifying [CMD: ``--recursive`` CMD][PY:
            ``recursive`` PY] or [CMD: ``--recursion-limit`` CMD][PY:
            ``recursion_limit`` PY]. For example, datasets located underneath
            the hidden folder `.datalad` will be reported even if [CMD:
            ``--include-hidden`` CMD][PY: ``include_hidden`` PY] is omitted.
            """,
            action='store_true'),
    )

    _validator_ = EnsureCommandParameterization(dict(
        path=EnsurePath(),
        depth=EnsureInt() & EnsureRange(min=0),
        recursive=EnsureBool(),
        recursion_limit=EnsureInt() & EnsureRange(min=0),
        include_files=EnsureBool(),
        include_hidden=EnsureBool(),
    ))

    _examples_ = [
        dict(text="Show up to 3 levels of subdirectories below the current "
                  "directory, including files and hidden contents",
             code_py="tree(depth=3, include_files=True, include_hidden=True)",
             code_cmd="datalad tree -L 3 --include-files --include-hidden"),
        dict(text="Find all top-level datasets located anywhere under ``/tmp``",
             code_py="tree('/tmp', recursion_limit=0)",
             code_cmd="datalad tree /tmp -R 0"),
        dict(text="Report all subdatasets recursively and their directory "
                  "contents, up to 1 subdirectory deep within each "
                  "dataset",
             code_py="tree(recursive=True, depth=1)",
             code_cmd="datalad tree -r -L 1"),
    ]

    @staticmethod
    @datasetmethod(name='tree')
    @eval_results
    def __call__(
            path='.',
            *,
            depth=None,
            recursive=False,
            recursion_limit=None,
            include_files=False,
            include_hidden=False):

        if recursive or recursion_limit is not None:
            # special tree defined by subdataset nesting depth
            tree_cls = DatasetTree
            dataset_tree_args = {"max_dataset_depth": recursion_limit}
        else:
            # simple tree defined by directory depth
            tree_cls = Tree
            dataset_tree_args = {}

        tree = tree_cls(
            path,
            max_depth=depth,
            exclude_node_func=build_excluded_node_func(
                include_hidden=include_hidden, include_files=include_files),
            **dataset_tree_args
        )

        for node in tree.generate_nodes():
            # yield one node at a time to improve UX / perceived speed
            res_dict = {
                "action": "tree",
                "path": str(node.path),
                "type": node.TYPE,
                "depth": node.depth,
                "exhausted_levels": list(tree.exhausted_levels),
                "count": {
                    "datasets": tree.node_count["DatasetNode"],
                    "directories": tree.node_count["DirectoryNode"],
                    **({"files": tree.node_count["FileNode"]}
                       if include_files else {})
                },
            }
            if node.TYPE == "dataset":
                res_dict.update({
                    "dataset_depth": node.ds_depth,
                    "dataset_abs_depth": node.ds_absolute_depth,
                    "dataset_is_installed": node.is_installed
                })

            if node.is_symlink():
                # TODO: should we inform if the symlink is recursive (as per
                #  `tree.is_recursive_symlink()`) although not broken? The
                #  UNIX 'tree' command shows the message '[recursive,
                #  not followed]' next to the path. Not sure if this is
                #  interesting at all or more confusing.
                res_dict["symlink_target"] = node.get_symlink_target()
                res_dict["is_broken_symlink"] = node.is_broken_symlink()

            if node.exception is not None:
                # mimic error message of unix 'tree' command for
                # permission denied error, otherwise use exception short
                # message
                message = "error opening dir" \
                    if node.exception.name == "PermissionError" \
                    else node.exception.message

                yield get_status_dict(
                    status="error",
                    message=message,
                    exception=node.exception,
                    **res_dict
                )
            else:
                yield get_status_dict(
                    status="ok",
                    **res_dict
                )

    @staticmethod
    def custom_result_renderer(res, **kwargs):
        """
        Each node is printed on one line. The string uses the format::

            [<indentation>] [<branch_tip_symbol>] [<ds_marker>] <path>

        Example line::

            │   │   ├── path_dir_level3
        """
        from datalad_next.uis import ansi_colors

        # get values from result record
        node_type = res["type"]
        node_path = res["path"]
        depth = res["depth"]
        exhausted_levels = res["exhausted_levels"]

        # build indentation string
        indentation = ""
        if depth > 0:
            indentation_symbols_for_levels = [
                ("│"
                 if level not in exhausted_levels
                 else " ") + "   "
                for level in range(1, depth)
            ]
            indentation = "".join(indentation_symbols_for_levels)

        # build prefix (tree branch tip)
        prefix = ""
        if depth > 0:  # root node has no prefix
            is_last_child = depth in exhausted_levels
            prefix = "└──" if is_last_child else "├──"

        # build dataset marker if dataset
        ds_marker = ""
        if node_type == "dataset":
            ds_absolute_depth = res["dataset_abs_depth"]
            ds_is_installed = res["dataset_is_installed"]

            ds_marker_depth = ansi_colors.color_word(
                f"DS~{ds_absolute_depth}",
                ansi_colors.WHITE)
            install_flag = " (not installed)" if not ds_is_installed else ""
            ds_marker = f"[{ds_marker_depth}]" + install_flag

        # build path string with optional color
        # display only root directory with full path, all other nodes
        # with basename
        path = node_path if depth == 0 else Path(node_path).name
        color_for_type = {
            "dataset": ansi_colors.MAGENTA,
            "directory": ansi_colors.BLUE,
            "file": None,
            "broken_symlink": ansi_colors.RED
        }
        # ANSI color for the path, if terminal colors are enabled
        color = color_for_type[node_type]
        if color is not None:
            path = ansi_colors.color_word(path, color)
        if res.get("is_broken_symlink", False):
            path = ansi_colors.color_word(path,
                                          color_for_type["broken_symlink"])

        # set suffix for directories
        dir_suffix = ""
        if depth > 0 and node_type in ("directory", "dataset"):
            dir_suffix = "/"

        # append symlink target if symlink
        symlink_target = ""
        if "symlink_target" in res:
            symlink_target = " -> " + res["symlink_target"]

        # add short error message if there was exception
        error_msg = ""
        if "exception" in res:
            error_msg = f" [{res['message']}]"

        line = indentation + \
            " ".join((s for s in (prefix, ds_marker, path) if s != "")) + \
            dir_suffix + symlink_target + error_msg
        ui.message(line)

    @staticmethod
    def custom_result_summary_renderer(res, **kwargs):
        """Print the summary 'report line' with count of nodes by type"""

        c_ds = res[-1]['count']['datasets']
        c_dirs = res[-1]['count']['directories']
        # files may not be included in results (if not using command
        # option '--include-files')
        c_files = res[-1]['count'].get('files')

        descriptions = [
            f"{c_ds} " + ("dataset" if int(c_ds) == 1 else "datasets"),
            f"{c_dirs} " + ("directory" if int(c_dirs) == 1 else "directories")
        ]
        if c_files is not None:
            descriptions.append(
                f"{c_files} " + ("file" if int(c_files) == 1 else "files"))

        ui.message("\n" + ", ".join(descriptions))


def build_excluded_node_func(include_hidden=False, include_files=False):
    """Return a function to exclude ``_TreeNode`` objects from the tree
    (prevents them from being yielded by the node generator).

    Returns
    -------
    Callable
        Function that takes the Path object of a ``_TreeNode`` as input,
        and returns true if the node should *not* be displayed in the tree.
    """

    def is_excluded(node: _TreeNode):
        return any((
            isinstance(node, FileNode) if not include_files else False,
            node.path.name.startswith(".") if not include_hidden else False
        ))

    return is_excluded


def increment_node_count(node_generator_func):
    """Decorator for incrementing the node count whenever a ``_TreeNode`` is
    yielded.

    Parameters
    ----------
    node_generator_func: Callable
        Function that yields ``_TreeNode`` objects
    """
    @wraps(node_generator_func)
    def _wrapper(*args, **kwargs):
        self = args[0]   # 'self' is a Tree instance
        for node in node_generator_func(*args, **kwargs):
            node_type = node.__class__.__name__
            if node_type not in self.node_count:
                raise ValueError(
                    f"No counts collected for unknown node type '{node_type}'"
                )
            if node.depth > 0:  # do not count the root directory
                # TODO: do not count symlinks if they point to
                #  files/directories that are already included in the tree
                #  (to prevent double counting)? Note that UNIX 'tree' does
                #  count double.
                self.node_count[node_type] += 1

            yield node  # yield what the generator yielded

    return _wrapper


def yield_with_last_item(generator):
    """Takes a generator and yields for each item, the item itself and
    whether it is the last item in the sequence.

    Returns
    -------
    Tuple[bool, Any]
        A tuple (is_last_item, item)
    """
    prev_val = next(generator, None)
    if prev_val is not None:
        for current_val in generator:
            yield False, prev_val
            prev_val = current_val
        yield True, prev_val


def path_depth(path: Path, root: Path):
    """Calculate directory depth of a path relative to the given root.

    Can also be a negative integer if the path is a parent of the
    tree root.

    Returns
    -------
    int
        Number of levels of the given path *below* the root (positive
        integer) or *above* the tree root (negative integer)

    Raises
    ------
    ValueError
        Like ``path.relative_to()``, raises ``ValueError`` if the path is not
        relative to the root
    """
    sign = 1
    try:
        rpath = path.relative_to(root)
    except ValueError:
        try:
            rpath = root.relative_to(path)
            sign = -1
        except ValueError:
            raise ValueError(
                "Could not calculate directory depth: "
                f"'{path}' is not relative to the tree root "
                f"'{root}' (or vice-versa)")
    return sign * len(rpath.parts)


def is_empty_dir(path: Path):
    """Does not check that path is a directory (to avoid extra
    system calls)"""
    return not any(path.iterdir())


@lru_cache()
def is_dataset(path: Path, installed_only=False):
    """Fast dataset detection.

    Infer that a directory is a dataset if it is either:

    - installed, or
    - not installed, but has an installed superdatset (only if argument
      ``installed_only`` is False)

    Only consider datalad datasets, not plain git/git-annex repos.

    Symlinks pointing to datasets are not resolved, so will always return
    False for symlinks. This prevents potentially detecting duplicate datasets
    if the symlink and its target are both included in the tree.

    Results are cached because the check is somewhat expensive and may
    be run multiple times on the same path.

    Parameters
    ----------
    path: Path
        Path to directory to be identified as dataset or non-dataset

    installed_only: bool
        Whether to ignore datasets that are not installed
    """
    try:
        if path.is_symlink():
            # ignore symlinks even if pointing to datasets, otherwise we may
            # get duplicate counts of datasets
            lgr.debug("Path is a symlink, will not check if it points to a "
                      "dataset: %s", path)
            return False

        if (path / ".datalad" / "config").is_file():
            # could also query `ds.id`, but checking just for existence
            # of config file is quicker.
            return True

        # if it is not installed, check if it has an installed superdataset.
        # instead of querying ds.is_installed() (which checks if the
        # directory has the .git folder), we check if the directory
        # is empty (faster) -- as e.g. after a non-recursive `datalad clone`
        if not installed_only:
            if is_empty_dir(path):
                return get_superdataset(path) is not None

    except Exception as ex:
        # if anything fails (e.g. permission denied), we raise exception
        # instead of returning False. this can be caught and handled by the
        # caller.
        raise NoDatasetFound(f"Cannot determine if '{path.name}' is a "
                             f"dataset") from ex

    return False


@lru_cache()
def get_subds_paths(ds_path: Path):
    """Return paths of immediate subdatasets for a given dataset path."""
    # This is an expensive operation because it calls git to read the
    # submodules. Since we need to run it to (A) calculate dataset depth and
    # (B) detect non-installed datasets, we cache results, so that the list of
    # subdatasets is computed only once for each parent dataset.
    return [
        str(ds_path / sm.path)
        for sm in iter_submodules(ds_path)
    ]


@lru_cache()
def get_dataset_root_datalad_only(path: Path):
    """Get root of dataset containing a given path (datalad datasets only,
    not pure git/git-annex repo)

    Parameters
    ----------
    path: Path
        Path to file or directory

    Returns
    -------
    Path
    """
    ds_root = path
    while ds_root:
        potential_ds_root = get_dataset_root(str(ds_root))

        if potential_ds_root is None:
            return None  # we are not inside a dataset

        potential_ds_root = Path(potential_ds_root)
        if is_dataset(potential_ds_root, installed_only=True):
            return potential_ds_root  # it's a match

        # we go one directory higher and try again
        ds_root = (potential_ds_root / "..").resolve(strict=True)
    return ds_root


@lru_cache()
def get_superdataset(path: Path):
    """Reimplementation of ``Dataset.get_superdataset()`` to allow caching
    results of `ds.subdatasets()` (the most expensive operation).

    Parameters
    ----------
    path: Path
        Path to a dataset

    Returns
    -------
    Dataset or None
    """
    superds_path = None

    while path:
        parent_path = (path / "..").resolve(strict=True)

        sds_path_ = get_dataset_root_datalad_only(parent_path)
        if sds_path_ is None:
            # no more parents, use previous found
            break

        superds = Dataset(sds_path_)

        # test if path is registered subdataset of the parent
        if not str(path) in get_subds_paths(superds.pathobj):
            break

        # That was a good candidate
        superds_path = sds_path_
        path = parent_path
        break

    if superds_path is None:
        # None was found
        return None
    return Dataset(superds_path)


def is_path_relative_to(my_path: Path, other_path: Path):
    """Port of pathlib's ``Path.is_relative_to()`` (requires python3.9+)"""
    try:
        my_path.relative_to(other_path)
        return True
    except ValueError:
        return False


class Tree:
    """Main class for generating and serializing a directory tree"""

    def __init__(self,
                 root: Path,
                 max_depth=None,
                 exclude_node_func=None):
        """
        Parameters
        ----------
        root: Path
            Directory to be used as tree root
        max_depth: int or None
            Maximum directory depth for traversing the tree
        exclude_node_func: Callable or None
            Function to filter out tree nodes from the tree
        """
        try:
            root = Path(root)
            self.root = root.resolve(strict=False)
            assert self.root.is_dir(), f"path is not a directory: {self.root}"
        except (AssertionError, OSError) as ex:  # could be permission error
            raise ValueError(f"directory not found: '{root}'") from ex

        self.max_depth = max_depth
        if max_depth is not None and max_depth < 0:
            raise ValueError("max_depth must be >= 0")

        # set callable to exclude nodes from the tree, meaning they
        # will not be yielded by the node generator
        self.exclude_node_func = exclude_node_func or self.default_exclude_func

        # keep track of levels where the subtree is exhausted, i.e. we
        # have reached the last node of the current subtree.
        # this is needed for the custom results renderer, to display nodes
        # differently depending on whether they are the last child or not.
        self.exhausted_levels = set([])

        # store dict with count of nodes for each node type, similar to the
        # tree command's 'report line' at the end of the output.
        # the node types (subclasses of ``_TreeNode``) are mutually exclusive,
        # so the sum of their counts equals to the total node count.
        # does not count the root itself, only the contents below the root.
        self.node_count = {node_type.__name__: 0
                           for node_type in _TreeNode.__subclasses__()}

    def __repr__(self):
        return self.__class__.__name__ + \
               f"('{self.root}', max_depth={self.max_depth})"

    @staticmethod
    def default_exclude_func(node):
        """By default, exclude files and hidden directories from the tree"""
        return any(
            (isinstance(node, FileNode), node.path.name.startswith("."))
        )

    def path_depth(self, path: Path):
        return path_depth(path, self.root)

    def _generate_tree_nodes(self, dir_path: Path):
        """Recursively yield ``_TreeNode`` objects starting from ``dir_path``

        Parameters
        ----------
        dir_path: Path
            Directory from which to calculate the tree
        """
        # yield current directory/dataset node
        current_depth = self.path_depth(dir_path)
        current_node = Node(dir_path, current_depth)
        yield current_node

        # check that we are within max_depth levels
        # (None means unlimited depth)
        if self.max_depth is None or \
                current_depth < self.max_depth:

            if current_node.is_symlink() and \
                    current_node.is_recursive_symlink(self.max_depth):
                # if symlink points to directory that we may visit or may
                # have visited already, do not recurse into it
                lgr.debug("Symlink is potentially recursive, "
                          "will not traverse target directory: %s", dir_path)
                return

            if current_node.exception is not None:
                # if some exception occurred when instantiating the node
                # (missing permissions etc), do not recurse into directory
                lgr.debug("Node has exception, will not traverse directory: "
                          "%r", current_node)
                return

            # sort child nodes alphabetically
            # needs to be done *before* calling the exclusion function,
            # because the function may depend on sort order
            all_children = sorted(list(dir_path.iterdir()))
            child_depth = current_depth + 1

            # generator to apply exclusion filter
            def children():
                for child_path in all_children:
                    child_node = Node(child_path, child_depth)
                    if not self.exclude_node_func(child_node):
                        yield child_node

            # exclusion function could be expensive to compute, so we
            # use a generator for child nodes. however, we need to be able
            # to detect the last child node within each subtree (needed for
            # displaying special end-of-subtree prefix). so we wrap the
            # generator in another 'lookahead' generator to detect the last
            # item.
            for is_last_child, child in yield_with_last_item(children()):

                if is_last_child:  # last child of its subtree
                    self.exhausted_levels.add(child_depth)
                else:
                    self.exhausted_levels.discard(child_depth)

                # remove exhausted levels that are deeper than the
                # current depth (we don't need them anymore)
                levels = set(self.exhausted_levels)  # copy
                self.exhausted_levels.difference_update(
                    l for l in levels if l > child_depth
                )

                if isinstance(child, (DirectoryNode, DatasetNode)):
                    # recurse into subdirectories
                    yield from self._generate_tree_nodes(child.path)
                else:
                    # it's a file, just yield it
                    yield child

    @increment_node_count
    def generate_nodes(self):
        """
        Traverse a directory tree starting from the root path.
        Yields ``_TreeNode`` objects, each representing a directory or
        dataset or file. Nodes are traversed in depth-first order.

        Returns
        -------
        Generator[_TreeNode]
        """
        # because the node generator is recursive, we cannot directly
        # decorate it with `increment_node_count` (since it would count
        # twice whenever the function recurses).
        # so we decorate a separate function where we just yield from the
        # underlying generator.
        yield from self._generate_tree_nodes(self.root)


class DatasetTree(Tree):
    """
    ``DatasetTree`` is a ``Tree`` whose depth is determined primarily
    by the subdataset hierarchy level (parameter ``max_dataset_depth``).

    Here, ``max_depth`` can also be specified, but it refers to the
    depth of each dataset's content. If this depth is 0, only datasets
    are reported, without any files or subdirectories underneath.

    Because of the different semantics of the ``max_depth`` parameter,
    this class is implemented as a separate subclass of ``Tree``.
    """
    def __init__(self, *args, max_dataset_depth=None, **kwargs):
        super().__init__(*args, **kwargs)

        # default max_dataset_depth 'None' means unlimited subdataset deoth
        self.max_dataset_depth = max_dataset_depth
        if self.max_depth is None:
            # by default, do not include datasets' contents
            self.max_depth = 0

        # lazy initialization of list of datasets and their parents,
        # will be computed when generating nodes for the first time
        self.ds_nodes = []

    def __repr__(self):
        return self.__class__.__name__ + \
                f"('{self.root}', " \
                f"max_dataset_depth={self.max_dataset_depth}, " \
                f"max_depth={self.max_depth})"

    @increment_node_count
    def generate_nodes(self):
        # compute full list of dataset nodes and their parents upfront.
        # this requires an unlimited-depth tree traversal, so will
        # be the slowest operation
        if not self.ds_nodes:
            lgr.debug("Started computing dataset nodes for %r", self)
            self.ds_nodes = list(self.generate_dataset_nodes())
            lgr.debug("Finished computing dataset nodes for %r", self)

        if not self.ds_nodes:
            depth = 0  # no datasets to report on, just yield the root
        else:
            depth = max(node.depth for node in self.ds_nodes) + \
                    self.max_depth  # max levels below the deepest dataset

        tree = Tree(
            self.root,
            max_depth=depth,
            exclude_node_func=self.exclude_func,
        )
        # synchronize exhausted levels with the main tree
        self.exhausted_levels = tree.exhausted_levels

        yield from tree.generate_nodes()

    def generate_dataset_nodes(self):
        """
        Generator of dataset nodes and their parent directories starting
        from below the tree root and up to ``max_dataset_depth`` levels.

        The assumption is that (super)datasets could be located at any level
        of the directory tree. Therefore, this function does a full-depth
        tree traversal to discover datasets.

        Returns
        -------
        Generator[DirectoryNode or DatasetNode]
        """

        def is_excluded(n: _TreeNode):
            # assumption: we won't find datasets underneath the git folder
            return isinstance(n, FileNode) or \
                   (isinstance(n, DirectoryNode) and n.path.name == ".git")

        # keep track of traversed nodes
        # (needed to prevent yielding duplicates)
        visited = set([])

        ds_tree = Tree(
            self.root,
            max_depth=None,  # unlimited depth, datasets could be anywhere
            exclude_node_func=is_excluded,
        )
        nodes_below_root = ds_tree.generate_nodes()
        next(nodes_below_root)  # skip root node

        for node in nodes_below_root:
            # for each dataset node, yield its parents first, then
            # yield the dataset itself
            if isinstance(node, DatasetNode) and \
                    (self.max_dataset_depth is None or
                     node.ds_depth <= self.max_dataset_depth) and \
                    not self.exclude_node_func(node):

                # yield parent directories if not already done
                parents_below_root = node.parents[1:]  # first parent is root
                for par_depth, par_path in enumerate(parents_below_root):
                    parent = Node(par_path, par_depth)

                    if parent not in visited:
                        visited.add(parent)
                        yield parent

                visited.add(node)
                yield node

    def exclude_func(self, node):
        """Exclusion function for pruning the main tree"""
        include, exclude = False, True  # prevent headaches

        try:
            if node in self.ds_nodes:
                # we hit a dataset or the parent of a dataset
                return include

            # if `max_depth` is specified for returning dataset contents,
            # exclude non-dataset nodes below a dataset that have
            # depth (relative to parent dataset) > max_depth
            if self.max_depth > 0 and \
                    not isinstance(node, DatasetNode):

                # check that node is the child of a dataset
                ds_parent = self._find_closest_ds_parent(node)
                if ds_parent is not None:
                    rel_depth = node.depth - ds_parent.depth
                    exceeds_max_depth = rel_depth > self.max_depth
                    # also filter by the user-supplied
                    # exclusion logic in `exclude_node_func`
                    return exceeds_max_depth or \
                        self.exclude_node_func(node)

        except Exception as ex:
            CapturedException(ex, level=10)  # DEBUG level
            lgr.debug("Excluding node from tree because "
                      "an exception occurred while applying "
                      "exclusion filter: %r", node)

        return exclude  # exclude by default

    def _find_closest_ds_parent(self, node):
        ds_parent = None
        for parent_path in node.path.parents:  # bottom-up order
            ds_parent = next((n for n in self.ds_nodes
                              if n.path == parent_path and
                              isinstance(n, DatasetNode)), None)
            if ds_parent is not None:
                break

        return ds_parent


class _TreeNode:
    """Base class for a directory or file represented as a single tree node
    and printed as single line of the 'tree' output."""
    TYPE = None  # needed for command result dict

    def __init__(self, path: Path, depth: int,
                 exception: CapturedException | None = None):
        """
        Parameters
        ----------
        path: Path
            Path of the tree node
        depth: int
            Directory depth of the node within its tree
        exception: CapturedException
            Exception that may have occurred at validation/creation
        """
        self.path = path
        self.depth = depth
        self.exception = exception

    def __eq__(self, other):
        return self.path == other.path

    def __hash__(self):
        return hash(str(self.path))

    def __repr__(self):
        return f"{self.__class__.__name__}('{self.path}', depth={self.depth})"

    @property
    def tree_root(self) -> Path:
        """Calculate tree root path from node path and depth"""
        parents = self.parents
        return parents[0] if parents \
            else self.path  # we are the root

    @property
    # More accurate annotation only from PY3.9 onwards
    # def parents(self) -> list[Path]:
    def parents(self) -> list:
        """List of parent paths in top-down order beginning from the tree root.
        Assumes the node path to be already normalized.

        Returns
        -------
        List[Path]
        """
        parents_from_tree_root = []
        for depth, path in enumerate(self.path.parents):
            if depth >= self.depth:
                break
            parents_from_tree_root.append(path)

        return parents_from_tree_root[::-1]  # top-down order

    def is_symlink(self) -> bool:
        """Check if node path is a symlink"""
        try:
            if self.path.is_symlink():
                return True
        except Exception as ex:
            # could fail because of permission issues etc.
            # in which case we just default to False
            self.exception = CapturedException(ex, level=10)
            return False

    def get_symlink_target(self) -> str:
        """If node path is a symlink, get link target as string. Otherwise,
        return None. Does not check that target path exists."""
        try:
            if self.is_symlink():
                # use os.readlink() instead of Path.readlink() for
                # Python <3.9 compatibility
                return readlink(str(self.path))
        except Exception as ex:
            self.exception = CapturedException(ex, level=10)

    def is_broken_symlink(self) -> bool:
        """If node path is a symlink, check if it points to a nonexisting
        or inaccessible target or to itself (self-referencing link). Raise
        exception if the node path is not a symlink."""
        if not self.is_symlink():
            raise ValueError("Node path is not a symlink, cannot check if "
                             f"symlink is broken: {self.path}")

        try:
            self.path.resolve(strict=True)
            return False
        except FileNotFoundError:  # target does not exist
            return True
        except PermissionError:  # target exists but is not accessible
            return True
        except (RuntimeError, OSError):  # symlink loop (OSError on Windows)
            return True
        except Exception as ex:  # probably broken in some other way
            self.exception = CapturedException(ex, level=10)
            return True

    def is_recursive_symlink(self, max_depth) -> bool:
        """Detect symlink pointing to a directory within the same tree
        (directly or indirectly).

        The default behaviour is to follow symlinks when traversing the tree.
        However, we should not follow symlinks to directories that we may
        visit or have visited already, i.e. are also located under the tree
        root or any parent of the tree root (within a distance of
        ``max_depth``).

        Otherwise, the same subtree could be generated multiple times in
        different places, potentially in a recursive loop (e.g. if the
        symlink points to its parent).

        This is similar to the logic of the UNIX 'tree' command, but goes a
        step further to prune all duplicate subtrees.

        Parameters
        ----------
        max_depth
            Max depth of the ``Tree`` to which this node belongs
        """
        if not self.is_symlink():
            raise ValueError("Node path is not a symlink, cannot check if "
                             f"symlink is recursive: {self.path}")

        if isinstance(self, FileNode):
            # we are only interested in symlinks pointing to a directory
            return False

        if self.is_broken_symlink():
            # cannot identify target, no way to know if link is recursive
            return False

        target_dir = self.path.resolve()
        tree_root = self.tree_root

        # either:
        # - target dir is within `max_depth` levels beneath the tree
        #   root, so it will likely be yielded or has already been
        #   yielded (bar any exclusion filters)
        # - target dir is a parent of the tree root, so we may still
        #   get into a loop if we recurse more than `max_depth` levels
        try:
            rel_depth = abs(path_depth(target_dir, tree_root))
            return max_depth is None or \
                rel_depth <= max_depth
        except ValueError:
            # cannot compute path depth because target is outside
            # of the tree root, so no loop is possible
            return False


class Node:
    """
    Factory class for creating a ``_TreeNode`` of a particular subclass.
    Detects whether the path is a file or a directory or dataset,
    and handles any exceptions (permission errors, broken symlinks, etc.)
    """
    def __new__(cls, path: Path, depth: int, **kwargs):
        if not isinstance(path, Path):
            raise ValueError("path must be a Path object")

        node_cls = FileNode
        captured_ex = None
        try:
            if path.is_dir():
                if is_dataset(path):
                    node_cls = DatasetNode
                else:
                    node_cls = DirectoryNode
        except NoDatasetFound as ex:  # means 'is_dataset()' failed
            # default to directory node
            # just log the exception, do not set it as node attribute
            CapturedException(ex, level=10)
            node_cls = DirectoryNode
        except Exception as ex:  # means 'is_dir()' failed
            # default to file node
            # set exception as node attribute
            captured_ex = CapturedException(ex, level=10)

        return node_cls(path, depth, exception=captured_ex, **kwargs)


class DirectoryNode(_TreeNode):
    TYPE = "directory"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        try:
            # get first child if exists. this is a check for whether
            # we can potentially recurse into the directory or
            # if there are any filesystem issues (permissions errors, etc)
            any(self.path.iterdir())
        except OSError as ex:
            # permission errors etc. are logged and stored as node
            # attribute so they can be passed to results dict.
            # this will overwrite any exception passed to the constructor,
            # since we assume that this exception is closer to the root
            # cause.
            self.exception = CapturedException(ex, level=10)  # DEBUG level


class FileNode(_TreeNode):
    TYPE = "file"


class DatasetNode(_TreeNode):
    TYPE = "dataset"

    def __init__(self, *args, **kwargs):
        """Does not check if valid dataset. This needs to be done before
        creating the instance."""
        super().__init__(*args, **kwargs)

        try:
            self.ds = EnsureDataset(installed=None)(self.path).ds
            self.is_installed = self.ds.is_installed()
            self.ds_depth, self.ds_absolute_depth = self.calculate_dataset_depth()
        except Exception as ex:
            if self.exception is not None:
                # only if exception has not already been passed to constructor
                self.exception = CapturedException(ex, level=10)

    @lru_cache()
    def calculate_dataset_depth(self):
        """
        Calculate 2 measures of a dataset's nesting depth/level:

        1. ``ds_depth``: subdataset depth relative to the tree root
        2. ``ds_absolute_depth``: absolute subdataset depth in the full
           hierarchy, potentially taking into account parent datasets at
           levels above the tree root

        Returns
        -------
        Tuple[int, int]
            Tuple of relative dataset depth and absolute dataset depth
        """
        ds_depth = 0
        ds_absolute_depth = 0

        ds = self.ds
        while ds:
            superds = get_superdataset(ds.pathobj)

            if superds is None:
                # it is not a dataset, do nothing
                break
            else:
                if superds == ds:
                    # it is a top-level dataset, we are done
                    break

                ds_absolute_depth += 1
                if is_path_relative_to(superds.pathobj, self.tree_root):
                    # if the parent dataset is underneath the tree
                    # root, we increment the relative depth
                    ds_depth += 1

            ds = superds

        return ds_depth, ds_absolute_depth



"""Representations of DataLad datasets built on git/git-annex repositories

Two sets of repository abstractions are available :class:`LeanGitRepo` and
:class:`LeanAnnexRepo` vs. :class:`LegacyGitRepo` and :class:`LegacyAnnexRepo`.

:class:`LeanGitRepo` and :class:`LeanAnnexRepo` provide a more modern,
small-ish interface and represent the present standard API for low-level
repository operations. They are geared towards interacting with Git and
git-annex more directly, and are more suitable for generator-like
implementations, promoting low response latencies, and a leaner processing
footprint.

The ``Legacy*Repo`` classes provide a, now legacy, low-level API to repository
operations. This functionality stems from the earliest days of DataLad and
implements paradigms and behaviors that are no longer common to the rest of the
DataLad API. :class:`LegacyGitRepo` and :class:`LegacyAnnexRepo` should no
longer be used in new developments, and are not documented here.


.. currentmodule:: datalad_next.datasets
.. autosummary::
   :toctree: generated

   Dataset
   LeanGitRepo
   LeanAnnexRepo
   LegacyGitRepo
   LegacyAnnexRepo
"""

from datalad.distribution.dataset import (
    Dataset,
    # this does nothing but provide documentation
    # only kept here until this command is converted to
    # pre-call parameter validation
    # TODO REMOVE FOR V2.0
    EnsureDataset as NoOpEnsureDataset,
    # TODO REMOVE FOR V2.0
    datasetmethod,
    # TODO REMOVE FOR V2.0
    resolve_path,
)
from datalad.dataset.gitrepo import GitRepo as LeanGitRepo
from datalad.support.gitrepo import GitRepo as LegacyGitRepo

from datalad.support.gitrepo import GitRepo as LegacyGitRepo
from datalad.support.annexrepo import AnnexRepo as LegacyAnnexRepo

from .annexrepo import LeanAnnexRepo



"""Handling of Git's pathspecs with subdirectory mangling support

.. deprecated:: 1.6
   This module is deprecated. It has been migrated to the `datasalad library
   <https://pypi.org/project/datasalad>`__. Imports should be adjusted to
   ``datasalad.gitpathspec``.
"""

__all__ = ['GitPathSpec', 'GitPathSpecs']

import warnings

from datasalad.gitpathspec import (
    GitPathSpec,
    GitPathSpecs,
)

warnings.warn(
    '`datalad_next.gitpathspec` has been migrated to the datasalad library, '
    'adjust imports to `datasalad.gitpathspec`',
    DeprecationWarning,
    stacklevel=1,
)
